"""
File Purpose: tools related to drawing (and maybe animating) 2D patches.
"""

import matplotlib.pyplot as plt
import matplotlib.patches as mpl_patches
import xarray as xr

from .movies import MoviePlotElement, MoviePlotNode
from .plot_settings import PlotSettings
from .plot_tools import plt_transformer
from ..errors import (
    InputError, InputMissingError, InputConflictError,
    PlottingAmbiguityError, DimensionalityError,
)
from ..tools import (
    alias, alias_key_of, alias_child, simple_property,
    UNSET,
    pcAccessor,
    xarray_fill_coords, xarray_ensure_dims,
)


''' --------------------- PatchPlotElement & XarrayPatch --------------------- '''

@PlotSettings.format_docstring()
class PatchPlotElement(MoviePlotElement):
    '''plot a 2D Patch.
    Base class for useable 2D patch plot elements, e.g. XarrayRectanglePatchPlotElement.

    params: dict
        dict of patch-related parameters.
        Many are probably settable directly via patch.set(**params).
        E.g. 'width' and 'height' are options for rectangle patch.
    transform: {transform}

    cls.MPL_SETTERS: dict of {{param key: how to set that param}}
        for params with nonstandard setters in matplotlib, tells the alternative to patch.set(k=value):
            {{str k: str v}} --> patch.set(v=value)
            {{str k: (str attr, str v)}} --> patch.attr(v=value)
        Examples:
            - for Annulus patch, need MPL_SETTERS['r'] = 'radii',
                because patch.set(r=value) isn't allowed; it's patch.set(radii=value) instead.
            - for Arrow patch, need MPL_SETTERS['x'] = ('set_data', 'x'),
                because patch.set(x=value) isn't allowed; it's patch.set_data(x=value) instead.
    '''
    patch_cls = NotImplemented  # <-- subclass should specify.
    MPL_SETTERS = {}  # <-- subclass should specify.
    PATCH_SETTINGS_LOOKUP = 'mpl.patches.Patch'  # <-- subclass should update.

    def __init__(self, params, *, ax=None, transform=UNSET, **kw_super):
        kw_super.update(transform=transform)
        super().__init__(**kw_super)
        self._ax_init = ax
        self.params = params
        self.init_patch()

    def _params_to_patch_init_kwargs(self, params):
        '''returns kwargs to use for initializing the patch, based on these params.
        The implementation here just returns params, unchanged.
        Subclass might override (e.g. RectanglePatchPlotElement join 'x0' and 'y0' into 'xy').
        The result should be a dict of kwargs suitable for self.patch_cls.__init__.
        '''
        return params

    def _params_to_patch_update_kwargs(self, params):
        '''returns kwargs to use for updating the patch, based on these params.
        The implementation here just returns params, unchanged.
        Subclass might override (e.g. RectanglePatchPlotElement rename 'x0' to 'x' and 'y0' to 'y').
        The result should be a dict of kwargs suitable for self._set_patch_params
            (i.e., before applying logic due to MPL_SETTERS).
        '''
        return params

    params = alias_key_of('data', 'params',
        doc='''dict of patch parameters. Internally, stored at self.data['params']''')
    patch = simple_property('_patch', doc='''the matplotlib.patches.Patch object''')
    ax = alias_child('patch', 'axes', if_no_child=None, doc='''the axes containing this patch''')
    fig = alias_child('patch', 'figure', if_no_child=None, doc='''the figure containing this patch''')

    def _all_patch_params(self, ax):
        '''returns full list of all known patch params to use, i.e.:
        {**params_settings, **params_init}, where:
            params_settings = self.plot_settings.get_mpl_kwargs(self.PATCH_SETTINGS_LOOKUP)
            params_init = self._params_to_patch_init_kwargs()
        '''
        params_settings = self.plot_settings.get_mpl_kwargs(self.PATCH_SETTINGS_LOOKUP)
        params_init = self._params_to_patch_init_kwargs(self.params)
        params_transform = self._transform_kw(ax=ax)
         # settings first in order below --> other params takes precedence if overlap.
        return {**params_settings, **params_init, **params_transform}

    def _transform_kw(self, ax):
        '''get transform to use for this patch, as dict.
        result is probably dict(transform=Transform object), but might be empty dict().
        requires that self.ax is not None.
        '''
        transform = self.plot_settings['transform']
        if transform is UNSET:
            return {}
        transform = plt_transformer(transform, ax=ax)
        return {'transform': transform}

    # # # PLOTTING THIS ELEMENT # # #
    def init_patch(self):
        '''initialize the patch; actually draw the patch on self.ax.
        stores plotted object in self.patch and returns self.patch.
        '''
        ax = self._ax_init
        if ax is None: ax = plt.gca()
        kw_patch = self._all_patch_params(ax=ax)
        self.patch = self.patch_cls(**kw_patch)
        ax.add_patch(self.patch)
        return self.patch

    # # # UPDATING (REQUIRED BY PARENT) # # #
    def update_data(self, data):
        '''updating the plot using data['params'].
        return the list of all updated matplotlib Artist objects (i.e., [self.patch])
        '''
        params = data['params']
        params = self._params_to_patch_update_kwargs(params)
        self._set_patch_params(**params)
        self.params = {**self.params, **params}  # bookkeeping
        return [self.patch]

    def _set_patch_params(self, **params):
        '''update the parameters of self.patch.

        params: dict
            dict of patch parameters.
            All must be provideable during (type of patch).__init__.
            Many are probably settable directly via patch.set(**params).
            E.g. 'width' and 'height' are options for rectangle patch.
            Some might not be settable directly; see self.MPL_SETTERS and help(self) for details.
        '''
        to_set = {}
        for k, value in params.items():
            if k in self.MPL_SETTERS:
                v = self.MPL_SETTERS[k]
                if isinstance(v, tuple):
                    attr, v = v
                    setter = getattr(self.patch, attr)
                    setter(**{v: value})
                else:
                    to_set[v] = value
            else:
                to_set[k] = value
        self.patch.set(**to_set)

    # # # CONENIENCE # # #
    def legend_handle(self, *, hatch_density_scaling=2, label=UNSET):
        '''returns a Patch suitable for use as a handle in a legend.
        
        hatch_density_scaling: int
            increase the hatch density by this factor, relative to the plot itself, if using hatches.
        label: UNSET or str
            if not UNSET, use this as the label, instead of self.plot_settings['label']

        Example:
            handle1 = self.legend_handle()
            handle2 = other_patch_element.legend_handle()
            plt.legend(handles=[handle1, handle2])
        '''
        kw = self.plot_settings.get_mpl_kwargs('mpl.patches.Patch', label=label)
        # intentionally hard-coded 'mpl.patches.Patch' here; legend doesn't care about Patch subclass params.
        if 'hatch' in kw:
            kw['hatch'] = kw['hatch'] * hatch_density_scaling
        return mpl_patches.Patch(**kw)


@PlotSettings.format_docstring()
class XarrayPatch(MoviePlotNode):
    '''MoviePlotNode of 2D patch.
    stores a PatchPlotElement * has methods for plotting & updating it.
    "patch" refers to a matplotlib.patches.Patch object.

    ds: xarray.Dataset, probably 0D or 1D.
        xarray Dataset containing the patch params to be plotted.
        if ndim=0, can still plot, but nothing to animate.
        patch pa
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None --> infer from ds.dims. (if ds is 1D, use t=ds.dims[0]).
    ax: None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.

    init_plot: {init_plot}
    label: {label}

    additional kwargs can contain constant patch params with same key as in matplotlib,
        but only if not provided in ds. E.g. 'edgecolor', but only if no ds['edgecolor'].
    additional kwargs can also be any other PlotSettings.
    '''
    element_cls = PatchPlotElement
    PATCH_SETTINGS_LOOKUP = 'pc.PatchPlotElement'

    def __init__(self, ds, t=None, ax=None,
                 init_plot=PlotSettings.default('init_plot'),  # <- could go in kw_super, but explicit is nice.
                 **kw_super):
        if isinstance(ds, xr.DataArray):
            ds = ds.to_dataset()
        ds = xarray_fill_coords(ds, need=[t])
        ds = xarray_ensure_dims(ds, ({t} - {None}), promote_dims_if_needed=True)
        self.ds = ds
        self._ax_init = ax
        if len(ds.dims) > 1:
            raise DimensionalityError(f'XarrayPatch expects 0D or 1D input, got dims: {ds.dims}')
        if t is None and len(ds.dims) == 1:
            t = list(ds.dims)[0]
        self.t = t
        # super init
        kw_super.update(init_plot=init_plot)
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    t_plot_dim = alias('t')

    ax = alias_child('obj', 'ax', if_special_child={None: None, UNSET: None},
        doc='''mpl.axes.Axes where this XarrayPatch is plotted, or None if not plotted.''')

    fig = alias_child('ax', 'figure', if_special_child={None: None, UNSET: None},
        doc='''mpl.figure.Figure where this XarrayPatch is plotted, or None if not plotted.''')

    patch = alias_child('obj', 'patch', if_special_child={None: None, UNSET: None},
        doc='''mpl.patches.Patch object of this XarrayPatch, or None if not plotted.''')

    legend_handle = alias_child('obj', 'legend_handle',
        doc='''alias to self.obj.legend_handle;
        method which returns a Patch suitable for use as a handle in a legend.
        (implied handles are sometimes okay, but this is useful to increase hatch density if using hatches.)''')

    # # # PLOTTING METHODS (REQUIRED BY PARENT CLASS) # # #
    def init_plot(self):
        '''plot for the first time. Save the PatchPlotElement in self.obj.'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs(self.PATCH_SETTINGS_LOOKUP)
        # ensure no overlap with data['params'].
        if any(k in kw_plot for k in data['params']):
            overlap = set(kw_plot.keys()) & set(data['params'].keys())
            raise PlottingAmbiguityError(f"some XarrayPatch kwargs also found in ds: {overlap}")
        # >> actually plot the patch <<
        self.obj = self.element_cls(data['params'], ax=self._ax_init, **kw_plot)
        # bookkeeping
        self.frame = frame

    def get_data_at_frame(self, frame):
        '''returns {'params': params at this frame}.'''
        t = self.t_plot_dim
        ds = self.ds
        if t is not None:
            ds = ds.isel({t: frame})
        result = {}
        for k, v in ds.items():  # loop instead of list comprehension, helps with debugging.
            result[k] = v.item()
        return {'params': result}

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        t = self.t_plot_dim
        if t is None:
            return 1
        return len(self.ds.coords[t])


''' --------------------- RectanglePatchPlotElement & XarrayRectanglePatch --------------------- '''

@PlotSettings.format_docstring()
class RectanglePatchPlotElement(PatchPlotElement):
    '''plot a rectangle patch.

    params: dict
        dict of any rectangle patch related parameters.
        must contain 'x0' and 'y0'.
        rectangle params are 'x0', 'y0', 'width', 'height'.
        (x0 and y0 point to lower left corner unless negative width or height.
            units determined by `transform`; default data units.)
    transform: {transform}

    'height' and 'width' may either be specified via `params` or as additional kwargs.
    Additional kwargs can be any patch params (with same key as in matplotlib) or PlotSettings.
    E.g. 'angle', 'facecolor', 'hatch'.

    See also: XarrayRectanglePatch, LimsPatchPlotElement
    '''
    patch_cls = mpl_patches.Rectangle
    _MPL_REQUIRED_PARAMS = ('xy', 'width', 'height')  # <-- this is just a note for bookkeeping purposes...
    PATCH_SETTINGS_LOOKUP = 'pc.RectanglePatchPlotElement'

    def _params_to_patch_init_kwargs(self, params):
        '''returns kwargs to use for initializing the patch, based on these params.
        The implementation here joins 'x' and 'y' into 'xy', but keeps other params unchanged.
        '''
        params = params.copy()
        x = params.pop('x0')
        y = params.pop('y0')
        return {'xy': (x, y), **params}

    def _params_to_patch_update_kwargs(self, params):
        '''returns kwargs to use for updating the patch, based on these params.
        The implementation here renames 'x0' and 'y0' to 'x' and 'y', but keeps other params unchanged.
        '''
        params = super()._params_to_patch_update_kwargs(params).copy()
        x = params.pop('x0')
        y = params.pop('y0')
        return {'x': x, 'y': y, **params}


@pcAccessor.register('rectangle_patch')
@PlotSettings.format_docstring()
class XarrayRectanglePatch(XarrayPatch):
    '''MoviePlotNode of 2D rectangle.
    stores a RectanglePatchPlotElement and has methods for plotting & updating it.

    ds: xarray.Dataset, probably 0D or 1D.
        xarray Dataset containing the rectangle params to be plotted.
        if ndim=0, can still plot, but nothing to animate.
        rectangle params are 'x0', 'y0', 'width', 'height'.
        (x0 and y0 point to lower left corner unless negative width or height.
            units determined by transform; default data units.)
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None --> infer from ds.dims. (if ds is 1D, use t=ds.dims[0]).
    ax: None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.
    transform: {transform}

    init_plot: {init_plot}
    label: {label}

    additional kwargs can contain constant rectangle params with same key as in matplotlib,
        but only if not provided in ds. E.g. 'edgecolor', but only if no ds['edgecolor'].
    additional kwargs can also be any other PlotSettings.

    See also: XarrayLimsPatch
    '''
    element_cls = RectanglePatchPlotElement
    PATCH_SETTINGS_LOOKUP = 'pc.RectanglePatchPlotElement'


''' --------------------- LimsPatchPlotElement & XarrayLimsPatch --------------------- '''


@PlotSettings.format_docstring()
class LimsPatchPlotElement(RectanglePatchPlotElement):
    '''plot a rectangle patch, from xmin, xmax, ymin, ymax, instead of x0, y0, width, height.

    params: dict
        dict of any rectangle patch related parameters.
        must contain ('xmin' and 'xmax') and/or ('ymin' and 'ymax').
        units determined by `transform`; default 'data' units.
        if xlims not provided, transform[0] must allow 'axes' units (will put 0 to 1, i.e. fill horizontally)
        if ylims not provided, transform[1] must allow 'axes' units (will put 0 to 1, i.e. fill vertically)
        (if transform restricted above but not input directly,
            default to 'axes' units where necessary, and 'data' units elsewhere.)
    transform: {transform}

    See also: XarrayLimsPatch, RectanglePatchPlotElement
    '''
    def _params_to_lims_and_transform(self, params):
        '''get (xmin, xmax, ymin, ymax, transform), from dict of params,
        and self.plot_settings['transform']
        '''
        transform = self.plot_settings.get('transform', params.get('transform', UNSET))
        if isinstance(transform, str):
            transform = (transform, transform)
        xmin = params.get('xmin', None)
        xmax = params.get('xmax', None)
        ymin = params.get('ymin', None)
        ymax = params.get('ymax', None)
        if (xmin is None) != (xmax is None):
            raise InputError(f'provide both xmin and xmax, or neither of them. Got xmin={xmin}, xmax={xmax}')
        if (ymin is None) != (ymax is None):
            raise InputError(f'provide both ymin and ymax, or neither of them. Got ymin={ymin}, ymax={ymax}')
        if xmin is None and ymin is None:
            raise InputMissingError('LimsPatchPlotElement requires at least one of xmin, xmax, ymin, ymax')
        if xmin is None:  # and ymin is not None
            if transform is UNSET:
                transform = ('axes', 'data')
            elif transform[0] != 'axes':
                raise InputConflictError('xmin and xmax not provided, but transform[0] is not "axes"')
            xmin = 0
            xmax = 1
        if ymin is None:
            if transform is UNSET:
                transform = ('data', 'axes')
            elif transform[1] != 'axes':
                raise InputConflictError('ymin and ymax not provided, but transform[1] is not "axes"')
            ymin = 0
            ymax = 1
        return (xmin, xmax, ymin, ymax, transform)

    def _params_to_patch_init_kwargs(self, params):
        '''returns kwargs to use for initializing the patch, based on self.params.
        The implementation converts xmin, xmax, ymin, ymax, into x0, y0, width, height,
            then calls super()._params_to_patch_init_kwargs() on the result.
        The implementation here also sets self.plot_settings['transform'].
        '''
        xmin, xmax, ymin, ymax, transform = self._params_to_lims_and_transform(params)
        width = xmax - xmin
        height = ymax - ymin
        x0 = xmin
        y0 = ymin
        params = params.copy()
        [params.pop(k, None) for k in ('xmin', 'xmax', 'ymin', 'ymax', 'transform')]
        params.update({'x0': x0, 'y0': y0, 'width': width, 'height': height})
        self.plot_settings['transform'] = transform
        return super()._params_to_patch_init_kwargs(params)

    def _params_to_patch_update_kwargs(self, params):
        '''returns kwargs to use for updating the patch, based on self.params.
        The implementation converts xmin, xmax, ymin, ymax, into x0, y0, width, height,
            then calls super()._params_to_patch_update_kwargs() on the result.
        '''
        xmin, xmax, ymin, ymax, _transform = self._params_to_lims_and_transform(params)
        width = xmax - xmin
        height = ymax - ymin
        x0 = xmin
        y0 = ymin
        params = params.copy()
        [params.pop(k, None) for k in ('xmin', 'xmax', 'ymin', 'ymax', 'transform')]
        params.update({'x0': x0, 'y0': y0, 'width': width, 'height': height})
        # <-- do NOT adjust params['transform'] here.
        return super()._params_to_patch_update_kwargs(params)


@pcAccessor.register('lims_patch')
@PlotSettings.format_docstring()
class XarrayLimsPatch(XarrayRectanglePatch):
    '''MoviePlotNode of 2D rectangle, from xmin, xmax, ymin, ymax, instead of x0, y0, width, height.
    stores a LimsPatchPlotElement and has methods for plotting & updating it.

    ds: xarray.Dataset, probably 0D or 1D.
        xarray Dataset containing the rectangle params to be plotted.
        must contain ('xmin' and 'xmax') and/or ('ymin' and 'ymax').
        units determined by `transform`; default 'data' units.
        if xlims not provided, transform[0] must allow 'axes' units (will put 0 to 1, i.e. fill horizontally)
        if ylims not provided, transform[1] must allow 'axes' units (will put 0 to 1, i.e. fill vertically)
        (if transform restricted above but not input directly,
            default to 'axes' units where necessary, and 'data' units elsewhere.)
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None --> infer from ds.dims. (if ds is 1D, use t=ds.dims[0]).
    ax: None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.
    transform: {transform}

    init_plot: {init_plot}
    label: {label}

    additional kwargs can contain constant rectangle params with same key as in matplotlib,
        but only if not provided in ds. E.g. 'edgecolor', but only if no ds['edgecolor'].
    additional kwargs can also be any other PlotSettings.

    See also: XarrayRectanglePatch

    Example:
        ds = xr.Dataset(dict(xmin=2.5, xmax=3.7))
        ds.pc.lims_patch(hatch='//', edgecolor='red', linewidth=0, facecolor='none')
        # produces a red hatched pattern between xmin=2.5 and xmax=3.7, with no outline and no fill, 
        #    extending from bottom to top of the axes.

        ds = xr.Dataset(dict(ymin=0.5, ymax=0.7), transform='axes')
        ds.pc.lims_patch(alpha=0.5)
        # produces a translucent rectangle between 50% and 70% of the way up the axes,
        #    extending from left to right of the axes.

        xx = pc.xr1d([2.5, 2.6, 2.7, 2.8, 2.9, 3.0], name='grid')
        ds = xr.Dataset(dict(xmin=xx, xmax=3.7))
        rect = ds.pc.lims_patch()
        rect.ani()  # or, xrect.save() to save to file instead of viewing in-line in Jupyter
        # produces an animation of rectangle gradually getting thinner (across 6 animation frames),
        #   spanning from x=2.5 to x=3.7 at first, but x=3.0 to x=3.7 by the end of the animation.
        # note: to attach this animation to an already existing PlasmaCalcs animation,
        #   e.g. xim=arr.image(), just add this MoviePlotNode to the tree, e.g. xim.add_child(rect).
    '''
    element_cls = LimsPatchPlotElement



"""
File Purpose: calculator for analyzing results of instability theory.
[TODO] this should probably go in the same place as addons.instability_tools code,
    however DimensionlessFromDatasetCalculator currently inherits from AddonsLoader.
    Would probably require refactoring AddonsLoader...
    (e.g., maybe subclasses of AddonsLoader should inject their known_vars
        at the same priority level as AddonsLoader known_vars?)
"""
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from .from_dataset import (
    MultifluidFromDatasetCalculator,
    VectorlessMultifluidFromDatasetCalculator,
)
from ..defaults import DEFAULTS
from ..dimensions import XHAT, YHAT, ZHAT
from ..plotting import cmap, title_from_coords
from ..quantities import magnitude
from ..errors import (
    LoadingNotImplementedError, InputConflictError,
    PlottingAmbiguityError,
)
from ..tools import (
    simple_property,
    xarray_min, xarray_drop_vars, xarray_sel,
    xarray_at_min_of, xarray_at_max_of,
)

''' --------------------- VectorlessInstabilityCalculator --------------------- '''

class VectorlessInstabilityCalculator(VectorlessMultifluidFromDatasetCalculator):
    '''calculator for analyzing results of instability theory.
    (Assumes itAccessor from instability_tools has been imported;
        some methods use syntax like, e.g., array.it.growth.)

    ds: xarray.Dataset
        should probably have all relevant base vars (& simple derived vars).
        Possibilities include:
            ds, m, q, gamma, n, mod_u, T, nusj, nusn, mod_E, mod_B, mod_E_un0, E_un0_perpmod_B,
            m_n or m_neutral, n_n or n_neutral, mod_u_n or mod_u_neutral, T_n or T_neutral.

        Probably should also have omega and k info.
    '''
    # # # CREATION OPTIONS # # #
    def __init__(self, ds, **kw_super):
        ds = self.assign_bases(ds)
        super().__init__(ds, **kw_super)

    @staticmethod
    def assign_bases(ds):
        '''return copy of ds, with some missing base vars assigned based on values in ds, if possible.
        Possibilities (only do it if all required values are present):
            q = skappa * m * nusn / mod_B
        '''
        if all([var in ds for var in ['m', 'mod_B', 'skappa', 'nusn']]):
            ds = ds.assign(q=ds['skappa'] * ds['m'] * ds['nusn'] / ds['mod_B'])
        return ds

    # # # VARS WHICH SHOULD COME FIRST # # #
    # (order matters. Check for these vars first.)

    @known_pattern(r'(.+)_at_kmax_of_(.+)', deps=[0,1])  # '{var}_at_kmax_of_{ref}'
    def get_var_at_kmax_of_ref(self, var, *, _match=None):
        '''var_at_kmax_of_ref --> self(var) at argmax of self(ref),
        taking argmax across all kdims in self(ref).
        For more precise control, consider directly using xarray_at_max_of.
        '''
        here, ref = _match.groups()
        refval = self(ref)
        return xarray_at_max_of(self(here), refval, refval.it.kdims)

    @known_pattern(r'(.+)_at_kmin_of_(.+)', deps=[0,1])  # '{var}_at_kmin_of_{ref}'
    def get_var_at_kmin_of_ref(self, var, *, _match=None):
        '''var_at_kmin_of_ref --> self(var) at argmin of self(ref),
        taking argmin across all kdims in self(ref).
        For more precise control, consider directly using xarray_at_min_of.
        '''
        here, ref = _match.groups()
        refval = self(ref)
        return xarray_at_min_of(self(here), refval, refval.it.kdims)

    # # # SPATIAL SCALE (need for timescales & compare to k) # # #
    cls_behavior_attrs.register('dspace_mode')
    DSPACE_MODE_OPTIONS = {
        'ds': '''self.load_direct('ds'). I.e. use self.ds['ds'].''',
        'k_at_growmax': '''infer from |k| at location (in k-space) of max growth. ds = 2 pi / |k|.''',
        'ldebye': '''infer from self('ldebye').''',
        'ldebye_total': '''infer from self('ldebye_total').''',
        'ldebye_minf': '''infer from self('ldebye'), taking min across all fluids.''',
        'eqperp_ldebye': '''infer from self('eqperp_ldebye').''',
        'eqperp_ldebye_total': '''infer from self('eqperp_ldebye_total').''',
        'eqperp_ldebye_minf': '''infer from self('eqperp_ldebye'), taking min across all fluids.''',
    }
    dspace_mode = simple_property('_dspace_mode',
        setdefaultvia='_default_dspace_mode', validate_from='DSPACE_MODE_OPTIONS',
        doc='''mode for calculating spatial scale self('ds') (required by speed-based timescales).
        See self.DSPACE_MODE_OPTIONS for options.''')
    def _default_dspace_mode(self):
        '''default dspace_mode. 'ds' if in self.ds, else 'k_at_growmax'.'''
        return 'ds' if 'ds' in self.ds else 'k_at_growmax'

    _DSPACE_MODE_TO_DEPS = {
        'ds': [],
        'k_at_growmax': ['kmod_at_growmax'],
        'ldebye': ['ldebye'],
        'ldebye_total': ['ldebye_total'],
        'ldebye_minf': ['ldebye'],
        'eqperp_ldebye': ['eqperp_ldebye'],
        'eqperp_ldebye_total': ['eqperp_ldebye_total'],
        'eqperp_minf_ldebye': ['eqperp_ldebye'],
    }
    @known_var(attr_deps=[('dspace_mode', '_DSPACE_MODE_TO_DEPS')], aliases=['dspace'])
    def get_ds(self):
        '''spatial scale. (required by self('timescales') for speed-based timescales.)
        result depends on self.dspace_mode. See self.DSPACE_MODE_OPTIONS for options.
        '''
        mode = self.dspace_mode
        if mode == 'ds':
            result = self.load_direct('ds')
        elif mode == 'k_at_growmax':
            result = 2 * np.pi / self('kmod_at_growmax')
        elif mode == 'ldebye':
            result = self('ldebye')
        elif mode == 'eqperp_ldebye':
            result = self('eqperp_ldebye')
        elif mode == 'ldebye_total':
            result = self('ldebye_total')
        elif mode == 'eqperp_ldebye_total':
            result = self('eqperp_ldebye_total')
        elif mode == 'ldebye_minf':
            result = xarray_min(self('ldebye'), 'fluid', missing_dims='ignore')
        elif mode == 'eqperp_ldebye_minf':
            result = xarray_min(self('eqperp_ldebye'), 'fluid', missing_dims='ignore')
        else:
            raise LoadingNotImplementedError(f'dspace_mode={mode!r} not recognized.')
        return result.assign_attrs(dspace_mode=mode)

    @known_var(deps=['ds'])
    def get_minf_ds(self):
        '''spatial scale, taking minimum across all fluids (i.e. most restrictive length scale).
        result depends on self.dspace_mode. See self.DSPACE_MODE_OPTIONS for options.
        '''
        return xarray_min(self('ds'), 'fluid', missing_dims='ignore')

    # # # GROWTH VARS # # #
    @known_var
    def get_omega(self):
        '''omega, directly from self.ds. (probably: roots of dispersion relation.)'''
        return self.load_direct('omega')

    @known_var(deps=['imag_omega'])
    def get_growth(self):
        '''imaginary part of omega.
        Equivalent: self('imag_omega'). Also equivalent: self('omega').it.growth
        '''
        return self('omega').it.growth

    @known_var(deps=['imag_omega'])
    def get_growth_kmax(self):
        '''imaginary part of omega, maxxed across k dims.
        Equivalent: self('omega').it.growth_kmax()
        Also equivalent: self('growth_at_growmax')
        '''
        return self('omega').it.growth_kmax()

    @known_var(deps=['imag_omega'])
    def get_grows(self):
        '''boolean array telling where self('growth_kmax')>0.
        Equivalent: self('omega').it.grows()
        '''
        return self('omega').it.grows()

    @known_pattern(r'(.+)_at_growmax', deps=[0, 'growth'])  # '{var}_at_growmax'
    def get_at_growmax(self, var, *, _match=None):
        '''value of var at location (in-kspace) of max growth.
        self('var_at_growmax') = self('var').it.at_growmax(self('growth'))

        Also equivalent: self('var_at_kmax_of_growth')
        '''
        # [TODO][EFF] slice before computing self(here), instead of after. 
        here, = _match.groups()
        return self(here).it.at_growmax(self('growth'))


    # [TODO][FIX] if trying to do '(.+)_at_growmax_where_(.+)' pattern,
    #   it gets picked up by '(.+)_where_(.+)' pattern first, for some reason.
    # workaround for now: use 'var_at_kmax_of_(growth_where_condition)'
    # @known_pattern(r'(.+)_at_growmax_where_(.+)', deps=[0, 1, 'growth'])
    # def get_at_growmax_where(self, var, *, _match=None):
    #     '''value of var at location (in-kspace) of max growth where condition is also True.
    #     Equivalent: self('var').it.at_growmax(self('growth').where(self('condition')))
    #     Equivalent: self('var_at_kmax_of_(growth_where_condition)')
    #     '''
    #     here, condition = _match.groups()
    #     return self(f'{var}_at_kmax_of_(growth_where_{condition})')

    @known_pattern(r'(.+)_within_kdebye', deps=[0, 'kmod', 'eqperp_ldebye'])
    def get_var_within_kdebye(self, var, *, _match=None):
        '''self(var) within |k| < eqperp_kdebye (==k corresponding to eqperp_ldebye).
        values at larger |k| are replaced with np.nan.
        '''
        here, = _match.groups()
        return self(f'{here}_where_(kmod<l2k_eqperp_ldebye)')

    @known_pattern(r'(.+)_within_kmfp', deps=[0, 'kmod', 'eqperp_lmfp'])
    def get_var_within_kmfp(self, var, *, _match=None):
        '''self(var) within |k| < eqperp_kmfp (==k corresponding to eqperp_lmfp).
        values at larger |k| are replaced with np.nan.
        '''
        here, = _match.groups()
        return self(f'{here}_where_(kmod<l2k_eqperp_lmfp)')

    # # # WAVEVECTOR VARS # # #
    @known_pattern(r'k2l_(.+)', deps=[0])  # k2l_var --> 2 * np.pi / var
    @known_pattern(r'l2k_(.+)', deps=[0])  # k2l_var --> 2 * np.pi / var
    def get_k2l(self, var, *, _match=None):
        '''convert from wavenumber to length, or vice versa.
        k2l_var = 2 pi / var.
        l2k_var = 2 pi / var.
        '''
        here, = _match.groups()
        return 2 * np.pi / self(here)

    @known_var
    def get_kmod(self):
        '''|k|, from self.ds kmod-related coord.
        (if self.ds tells 'log_kmod', returns 10**self.ds['log_kmod'].rename('kmod').
        else, returns self.ds['kmod'], unchanged.)
        Equivalent: self.ds.it.kmod
        '''
        return self.ds.it.kmod

    @known_var
    def get_kang(self):
        '''k angle, from self.ds kang-related coord.
        Equivalent: self.ds.it.kang
        '''
        return self.ds.it.kang

    @known_var(deps=['kang'])
    def get_khat(self):
        '''unit vector in direction of k.
        Roughly: self.ds.it.khat(), then sel(component=self.component)
        '''
        return xarray_sel(self.ds.it.khat(), component=self.component)

    @known_var(deps=['kmod', 'khat'])
    def get_k(self):
        '''k (vector). Inferred from kmod and kang.
        Roughly: self.ds.it.k(), then sel(component=self.component)
        Equivalent: self('kmod') * self('khat')
        '''
        return xarray_sel(self.ds.it.k(), component=self.component)

    # # # WAVE VELOCITY VARS # # #
    @known_var(deps=['real_omega', 'kmod'])
    def get_smod_vphase(self):
        '''signed magnitude of phase velocity: real(omega) / |k|.
        Phase velocity tells wave propagation speed. Group velocity tells wave envelope speed.
        Equivalent: self('omega').it.smod_vphase()
        See also: mod_vphase, vphase
        '''
        return self('omega').it.smod_vphase()

    @known_var(deps=['real_omega', 'kmod'], aliases=['mag_vphase'])
    def get_mod_vphase(self):
        '''magnitude of phase velocity: |real(omega) / |k||.
        Phase velocity tells wave propagation speed. Group velocity tells wave envelope speed.
        Equivalent: self('omega').it.mod_vphase()
        See also: smod_vphase, vphase
        '''
        return self('omega').it.mod_vphase()

    @known_var(deps=['real_omega', 'k'])
    def get_vphase(self):
        '''phase velocity (vector): vphase = (real(omega) / |k|) * khat.
        Phase velocity tells wave propagation speed. Group velocity tells wave envelope speed.
        Equivalent: self('omega').it.vphase()
        See also: smod_vphase, mod_vphase
        '''
        return self('omega').it.vphase()

    # # # GROWTH SPACESCALE VARS # # # 
    @known_var(deps=['kmod_at_growmax'])
    def get_spacescale_growth(self):
        '''length scale where growth is maximized: 2 pi / |k| at max growth.'''
        return 2 * np.pi / self('kmod_at_growmax')

    @known_var(deps=['spacescale_growth', 'minf_ds'])
    def get_spacesteps_growth(self):
        '''number of spacesteps (e.g. grid cells) needed to resolve length scale with max growth rate.
        Equivalent: self('spacescale_growth') / self('minf_ds').
        result depends on self.dspace_mode.

        Suggestion: if simulation grid cells have width ds, use N=safety * spacesteps_growth,
            with safety > 1 (e.g. safety = 10); safety will be approx. number of waves across the box.
        See also: self('ldebyesteps_growth'), self('eqperp_ldebyesteps_growth')
        '''
        return (self('spacescale_growth') / self('minf_ds')).assign_attrs(dspace_mode=self.dspace_mode)

    @known_var(deps=['spacescale_growth', 'ldebye'])
    def get_ldebyesteps_growth(self):
        '''number of Debye lengths needed to resolve length scale with max growth rate.
        Equivalent: self('spacescale_growth') / self('ldebye').
        Equivalent: self('spacesteps_growth', dspace_mode='ldebye').
        See also: self('spacesteps_growth'), self('eqperp_ldebyesteps_growth')
        '''
        return self('spacescale_growth') / self('ldebye')

    @known_var(deps=['spacescale_growth', 'eqperp_ldebye'])
    def get_eqperp_ldebyesteps_growth(self):
        '''number of eqperp Debye lengths needed to resolve length scale with max growth rate.
        Equivalent: self('spacescale_growth') / self('eqperp_ldebye').
        Equivalent: self('spacesteps_growth', dspace_mode='eqperp_ldebye').
        See also: self('spacesteps_growth'), self('ldebyesteps_growth')
        '''
        return self('spacescale_growth') / self('eqperp_ldebye')

    # # # GROWTH TIMESCALE VARS # # #
    TIMESCALE_VARS = VectorlessMultifluidFromDatasetCalculator.TIMESCALE_VARS.copy()
    if 'timescale_vtherm' in TIMESCALE_VARS:
        TIMESCALE_VARS.remove('timescale_vtherm')  # use eqperp_vtherm instead.
        TIMESCALE_VARS.append('timescale_eqperp_vtherm')
    TIMESCALE_VARS.extend(['timescale_vphase', 'timescale_growth'])

    @known_var(deps=['dsmin_for_timescales', 'mod_vphase_at_growmax'])
    def get_timescale_vphase(self):
        '''timescale from phase speed for waves at k with largest growthrate.
        timescale_vphase = dsmin / mod_vphase_at_growmax.
        '''
        return self('dsmin_for_timescales') / self('mod_vphase_at_growmax')

    @known_var(deps=['growth_kmax'])
    def get_timescale_growth(self):
        '''amount of time required for growth by a factor of e, at k with largest growthrate.
        Equivalent to 1/self('growth_kmax').
        Negative values indicate regions where growth rate is negative;
            consider using self('where_grows_timescale_growth') instead.
        '''
        return 1 / self('growth_kmax')

    @known_var(deps=['timescale_growth', 'minf_timescale'])
    def get_timesteps_growth(self):
        '''number of timesteps needed to see growth by a factor of e,
        assuming timestep = self('minf_timescale').
        Equivalent to self('timescale_growth') / self('minf_timescale').
        '''
        return self('timescale_growth') / self('minf_timescale')

    @known_var(deps=['timesteps_growth', 'spacesteps_growth'])
    def get_spacetimesteps_growth(self):
        '''number of "cells" across space AND time required to simulate growth. (Assumes 2D)
        Equivalent: self('spacesteps_growth')**2 * self('timesteps_growth').
        Result depends on self.dspace_mode.

        Note: if later applying safety factors to space, multiply result by safety**2.
            (e.g. might want to multiply spacesteps_growth 10 to see 10 waves in space, not just 1;
             to reflect that change in 'spacetimesteps_growth', need to multiply by 10**2 instead.)
        '''
        return self('spacesteps_growth')**2 * self('timesteps_growth')

    # # # NEW CALCULATOR VARS # # #
    BASES_CHECK = ['k', 'omega'] + list(VectorlessMultifluidFromDatasetCalculator.BASES_CHECK)
    # putting omega first ensures we keep all k dim info during self('bases').

    @known_var(deps=BASES_CHECK)
    def get_bases(self, **kw_get_vars):
        '''return dataset of all bases gettable based on self.ds.
        checks all vars from self.BASES_CHECK.
        '''
        return super().get_bases(**kw_get_vars)

    # # # ROSENBERG # # #
    @known_var(deps=['rosenberg_n_margin', 'ionnefrac_tiny'], ignores_dims='fluid')
    def get_rosenberg_n_scaler(self):
        '''return factor to DIVIDE all ion densities by, to push to rosenberg=1,
        for the ion with the most restrictive value for rosenberg criterion,
            ignoring all ions with tiny nnefrac.
            (i.e. ignore if n/ne < self.nnefrac_tiny_thresh. Default 0.01.)
        
        Equivalent: self('rosenberg_n_margin_where_not_nnefrac_tiny', fluid=self.fluids.ions()).min('fluid')

        densities scaling (of charged species) is "safe" for farley-buneman when rosenberg << 1,
            i.e. in that case it doesn't noticeably affect the physics.
        Suggestion: scale all (charged) densities by n_new = n_old / (safety * rosenberg_n_scaler),
            with safety less than 1, e.g. safety = 0.1.
        '''
        n_margins = self('rosenberg_n_margin_where_not_nnefrac_tiny', fluid=self.fluids.ions())
        return xarray_min(n_margins, 'fluid', missing_dims='ignore')

    # # # TITLE # # #
    default_title_width = simple_property('_default_title_width', default=80,
        doc='''default width for self._default_title(). Default: 80.''')

    title = simple_property('_title', setdefaultvia='_default_title',
        doc='''title for plots. Default: title_from_coords(self.ds, width=80)
        (uses self.default_title_width, which can also be edited if desired.)
        del self.title to reset to default.''')
    def _default_title(self):
        '''default title for plots. Uses title_from_coords.'''
        return title_from_coords(self.ds, width=self.default_title_width)


''' --------------------- InstabilityCalculator --------------------- '''

class InstabilityCalculator(VectorlessInstabilityCalculator,
                            MultifluidFromDatasetCalculator):
    '''instability calculator which also includes vector components.

    ds: xarray.Dataset
        should probably have all relevant base vars (& simple derived vars).
        Possibilities include:
            ds, m, q, gamma, n, u, T, nusj, nusn, E, B, E_un0,
            m_n or m_neutral, n_n or n_neutral, u_n or u_neutral, T_n or T_neutral.
    '''
    BASES_CHECK = ['k', 'omega'] + list(MultifluidFromDatasetCalculator.BASES_CHECK)

    @staticmethod
    def assign_bases(ds):
        '''return copy of ds, with some missing base vars assigned based on values in ds, if possible.
        Possibilities (only do it if all required values are present):
            q = skappa * m * nusn / mod_B   (will do mod(B) if B in ds but mod_B is not.)
            if component in ds and ds['component'] == ['Ehat', '-ExBhat']:
                replace 'Ehat' component with XHAT
                replace '-ExBhat' component with -YHAT
                B = ZHAT * mod_B
                E = XHAT * mod_E
                E_un0 = XHAT * E_un0_perpmod_B
            if component NOT in ds:
                B = ZHAT * mod_B
                E = XHAT * mod_E
                E_un0 = XHAT * E_un0_perpmod_B
            [TODO] further generalize component handling?
        '''
        if all([var in ds for var in ['m', 'mod_B', 'skappa', 'nusn']]):
            ds = ds.assign(q=ds['skappa'] * ds['m'] * ds['nusn'] / ds['mod_B'])
        elif all([var in ds for var in ['m', 'B', 'skappa', 'nusn']]):
            ds = ds.assign(q=ds['skappa'] * ds['m'] * ds['nusn'] / magnitude(ds['B']))
        # check if component needs to be converted for E & B...
        if 'component' in ds and ds['component'].size==2 and np.all(ds['component'] == ['Ehat', '-ExBhat']):
            to_assign = {}
            for varname, val in ds.data_vars.items():
                if 'component' in val.coords:
                    to_assign[varname] = val.isel(component=0)*XHAT + val.isel(component=1)*(-YHAT)
            ds = xarray_drop_vars(ds, to_assign, drop_unused_dims=True)  # make sure the old 'component' dim goes away.
            ds = ds.assign(to_assign)
            if 'mod_B' in ds:
                B = ZHAT * ds['mod_B']
                if 'B' in ds and not np.allclose(ds['B'], B):
                    raise InputConflictError('B conflicts with mod_B')
                ds = ds.assign(B=B)
            if 'mod_E' in ds:
                E = XHAT * ds['mod_E']
                if 'E' in ds and not np.allclose(ds['E'], E):  # [TODO] encapsulte instead of repeating...
                    raise InputConflictError('E conflicts with mod_E')
                ds = ds.assign(E=E)
            if 'E_un0_perpmod_B' in ds:
                E_un0 = XHAT * ds['E_un0_perpmod_B']
                if 'E_un0' in ds and not np.allclose(ds['E_un0'], E_un0):
                    raise InputConflictError('E_un0 conflicts with E_un0_perpmod_B')
                ds = ds.assign(E_un0=E_un0)
        elif 'component' not in ds:
            if 'mod_B' in ds:
                B = ZHAT * ds['mod_B']
                if 'B' in ds and not np.allclose(ds['B'], B):
                    raise InputConflictError('B conflicts with mod_B')
                ds = ds.assign(B=B)
            if 'mod_E' in ds:
                E = XHAT * ds['mod_E']
                if 'E' in ds and not np.allclose(ds['E'], E):
                    raise InputConflictError('E conflicts with mod_E')
                ds = ds.assign(E=E)
            if 'E_un0_perpmod_B' in ds:
                E_un0 = XHAT * ds['E_un0_perpmod_B']
                if 'E_un0' in ds and not np.allclose(ds['E_un0'], E_un0):
                    raise InputConflictError('E_un0 conflicts with E_un0_perpmod_B')
                ds = ds.assign(E_un0=E_un0)
        return ds

    # # # PLOTTERS # # #
    NMUL_TIMELINES_STYLE = {
        'cycles': [{
            'color': cmap('tab10').colors[:5],
            'markersize': [18,15,12,10,8],
            'markeredgewidth': [2,1.5,1.5,1,1],
        }],
    }

    @known_plotter(kinds=['timelines', 'k', 'at_growmax', 'n_mul'], cost=11)
    def plot_kmod_at_growmax(ec, **kw_plot_settings):
        '''timelines of kmod_at_growmax vs n_mul. (log-log plot.)
        Also includes lines for each fluid, for:
            kmod_at_growmax with kmod limited by kmfp or kdebye, i.e.:
                kmod_at_kmax_of_growth_within_eqperp_kmfp,
                kmod_at_kmax_of_growth_within_eqperp_ldebye,
            eqperp_kmfp and eqperp_ldebye, i.e.:
                l2k_eqperp_lmfp,
                l2k_eqperp_ldebye,
        '''
        plt.figure(figsize=(9, 6))
        arr = ec('log10_kmod_at_growmax')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')        
        n_mul = arr.coords['n_mul']
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, label='(full)',
                         color='black', lw=8, alpha=0.2,
                         **kw_plot_settings)
        plt.grid()
        # kmfp limited and kdebye limited
        arr = ec('log10_kmod_at_kmax_of_growth_within_kdebye')
        arr.pc.timelines(tCOORD, label='(kdebye_limited)',
                         marker='X', fillstyle='none',
                         **{**ec.NMUL_TIMELINES_STYLE, **kw_plot_settings})
        arr = ec('log10_kmod_at_kmax_of_growth_within_kmfp')
        arr.pc.timelines(tCOORD, label='(kmfp_limited)',
                         marker='o', fillstyle='none',
                         **{**ec.NMUL_TIMELINES_STYLE, **kw_plot_settings})
        arr = ec('log10_l2k_eqperp_lmfp')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            arr = arr * xr.ones_like(n_mul)
        arr.pc.timelines(tCOORD, label='(mfp)',
                         cycles=[DEFAULTS.PLOT.TIMELINES_CYCLE1],
                         color='blue', lw=2, alpha=0.3, zorder=-5,
                         **kw_plot_settings)

        arr = ec('log10_l2k_eqperp_ldebye')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            arr = arr * xr.ones_like(n_mul)
        arr.pc.timelines(tCOORD, label='(ldebye)',
                         cycles=[DEFAULTS.PLOT.TIMELINES_CYCLE1],
                         color='red', lw=2, alpha=0.3, zorder=-5,
                         **kw_plot_settings)
        # better labels
        plt.title(f'log10_kmod [{ec.units}]\n{ec.title}')
        plt.ylabel(f'log10_kmod_at_growmax [{ec.units}]')
        plt.xlabel('log10_n_mul')
        return result

    @known_plotter(kinds=['timelines', 'k', 'at_growmax', 'n_mul'], cost=10)
    def plot_kang_at_growmax(ec, **kw_plot_settings):
        '''timelines of kang_at_growmax vs n_mul. (using log of n_mul)
        Also includes lines for each fluid, for:
            kang_at_growmax with kmod limited by kmfp or kdebye, i.e.:
                kang_at_kmax_of_growth_within_eqperp_kmfp,
                kang_at_kmax_of_growth_within_eqperp_ldebye,
        '''
        plt.figure(figsize=(9, 4))
        arr = ec('rad2deg_kang_at_growmax')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, label='(full)',
                         color='black', lw=8, alpha=0.2,
                         **kw_plot_settings)
        plt.grid()
        # kmfp limited and kdebye limited
        arr = ec('rad2deg_kang_at_kmax_of_growth_within_kdebye')
        arr.pc.timelines(tCOORD, label='(kdebye_limited)',
                         marker='X', fillstyle='none',
                         **{**ec.NMUL_TIMELINES_STYLE, **kw_plot_settings})
        arr = ec('rad2deg_kang_at_kmax_of_growth_within_kmfp')
        arr.pc.timelines(tCOORD, label='(kmfp_limited)',
                         marker='o', fillstyle='none',
                         **{**ec.NMUL_TIMELINES_STYLE, **kw_plot_settings})
        # better labels
        plt.title(f'kang [deg]\n{ec.title}')
        plt.ylabel('kang_at_growmax [deg]')
        plt.xlabel('log10_n_mul')
        return result

    @known_plotter(kinds=['timelines', 'growth', 'at_growmax', 'n_mul'], cost=10)
    def plot_growth_at_growmax(ec, **kw_plot_settings):
        '''timelines of growth_at_growmax vs n_mul. (using log of n_mul)
        Also includes lines for each fluid, for:
            growth_at_growmax with kmod limited by kmfp or kdebye, i.e.:
                growth_at_kmax_of_growth_within_eqperp_kmfp,
                growth_at_kmax_of_growth_within_eqperp_ldebye,
        '''
        plt.figure(figsize=(9, 5))
        arr = ec('log10_growth_at_growmax')
        if 'n_mul' not in arr.coords and 'log_n_mul' not in arr.coords:
            raise PlottingAmbiguityError('expected n_mul dependence')
        tCOORD = 'log_n_mul'
        result = arr.pc.timelines(tCOORD, label='(full)',
                         color='black', lw=8, alpha=0.2,
                         **kw_plot_settings)
        plt.grid()
        # kmfp limited and kdebye limited
        arr = ec('log10_growth_at_kmax_of_growth_within_kdebye')
        arr.pc.timelines(tCOORD, label='(kdebye_limited)',
                         marker='X', fillstyle='none',
                         **{**ec.NMUL_TIMELINES_STYLE, **kw_plot_settings})
        arr = ec('log10_growth_at_kmax_of_growth_within_kmfp')
        arr.pc.timelines(tCOORD, label='(kmfp_limited)',
                         marker='o', fillstyle='none',
                         **{**ec.NMUL_TIMELINES_STYLE, **kw_plot_settings})
        # better labels
        plt.title(f'log10_growth [{ec.units}]\n{ec.title}')
        plt.ylabel(f'log10_growth_at_growmax [{ec.units}]')
        plt.xlabel('log10_n_mul')
        return result

    @known_plotter(kinds=['2D', 'subplots', 'growth', 'n_mul'], cost=21)
    def plot_growthplots(ec, *, share_vlims='row', **kw_plot_settings):
        '''ec.ds.it.growthplots(row='n_mul', wrap=6, **kw_plot_settings).
        Sets |k| plot lims to be the same on each plot,
            which is useful e.g. if |k| lims in ec.ds is based on kdebye, which varies with n_mul,
            e.g. from tfbi_solver(kres='mid', mod=dict(min=2*np.pi/1e2, max=(1.01, 'kdebye'))).solve()
        '''
        defaults = {
            'row': 'n_mul',
            'wrap': 6,
            'share_vlims': share_vlims,
            'suptitle_width': ec.default_title_width,
            'top': 1.1,  # fudge factor for prettier subplot spacing
        }
        log_kmod = ec.ds['log_kmod'] if 'log_kmod' in ec.ds else np.log10(ec.ds['kmod'])
        klims = [log_kmod.min().item(), log_kmod.max().item()]
        kw = {**defaults, **kw_plot_settings}
        return ec.ds.it.growthplots(y='log_kmod', ylim=klims, **kw)



"""
File Purpose: DimensionlessPlasmaCalculatorFromDataset
"""

import numpy as np
import xarray as xr

from ..dimensions import SnapList, ComponentList, FluidList
from ..errors import LoadingNotImplementedError
from ..plasma_calculator import (
    DimensionlessPlasmaCalculator,
    ComponentHavingPlasmaCalculator, PlasmaCalculator, MultifluidPlasmaCalculator,
    VectorlessPlasmaCalculator, VectorlessMultifluidPlasmaCalculator,
)
from ..quantities import DirectBasesLoader
from ..tools import (
    alias,
    is_iterable, xr1d,
    xarray_sel,
)

class DimensionlessFromDatasetCalculator(DirectBasesLoader, DimensionlessPlasmaCalculator):
    '''PlasmaCalculator taking a Dataset as input.
    No smart behavior implemented here about dimensions,
        e.g. getting quantities depending on vector arithmetic or certain fluids will fail.

    ds: xarray.Dataset
        self(var) looks in self.ds when can't find another way to compute var.
        For base vars, always checks ds first.
        (base vars): ds, q, gamma, mod_E, mod_B, E_perpmod_B, E_un0_perpmod_B,
                     m, n, mod_u, T, nusj,
                     m_neutral, n_neutral, mod_u_neutral, T_neutral.
        For simple derived vars, check ds first; compute from base vars if missing.
        (simple derived vars): r, p, P, Tjoule, e, nusn, nq, Jf, J, E_un0.
    '''
    # bases checked in self('bases')
    BASES_CHECK = ['ds', 'q', 'gamma', 'mod_E', 'mod_B', 'E_perpmod_B',
                    'm', 'n', 'mod_u', 'T', 'nusj',
                    'm_neutral', 'n_neutral', 'mod_u_neutral', 'T_neutral',
                    'nusn', 'E_un0_perpmod_B',
                    ]

    def __init__(self, ds, **kw_super):
        self.ds = ds
        super().__init__(**kw_super)

    def directly_loadable_vars(self):
        '''returns directly loadable variables from self.ds'''
        return list(self.ds.coords) + list(self.ds.data_vars)

    def load_fromfile(self, var, *args, **kw):
        '''load var from self.ds.'''
        if var in self.ds:
            return self.ds[var]
        else:
            raise LoadingNotImplementedError(f'var={var!r} not found in self.ds.')

    # # # KNOWN VARS / PATTERNS # # #
    @known_pattern(r'(mod|mag)2_(.+)', deps=[1])
    def get_mod2(self, var, *, _match=None):
        '''square of a variable. mod2_var = mod_var**2.
        mag2_var = mag_var**2.
        '''
        # [TODO] should this be defined elsewhere? E.g. in DimensionlessPlasmaCalculator...
        # (it's only necessary to understand mod2_var syntax without knowing vector arithmetic.)
        mod, here = _match.groups()
        return self(f'{mod}_{here}')**2

    @known_var(deps=BASES_CHECK)
    def get_bases(self, **kw_get_vars):
        '''return dataset of all bases gettable based on self.ds.
        checks all vars from self.BASES_CHECK.
        '''
        bases = self.BASES_CHECK
        kw_get_vars.setdefault('missing_vars', 'ignore')  # it's okay if some bases are un-gettable.
        return self.get_vars(bases, **kw_get_vars)

    # # # DISPLAY # # #
    def __repr__(self):
        ds_sizes = dict(self.ds.sizes)
        return f'{type(self).__name__}(ds={type(self.ds).__name__} with sizes={ds_sizes}))'

    # # # MISC. CONVENIENT METHODS # # #
    def with_isel(self, isel_dict, **kw_init):
        '''returns new instance of type(self), initialized with self.ds.isel(isel_dict).
        Does NOT maintain behavior attrs from self.
        Equivalent: type(self)(self.ds.isel(isel_dict), **kw_init)
        '''
        return type(self)(self.ds.isel(isel_dict), **kw_init)

    def with_scaling(self, scaling_dict, **kw_init):
        '''returns new instance of type(self), scaling some vars from self.ds.
        Does NOT maintain behavior attrs from self.
        result is type(self)(scaled_ds, **kw_init), where
            scaled_ds = self.ds.assign({v: scaling_dict[v] * self.ds[v] for v in scaling_dict}).

        scaling_dict: dict
            keys should be data_var names.
            vals should be numbers, 1D arraylike of numbers, or DataArray.
            if 1D non-DataArray arraylike, replace by xr1d(np.array(val), name='{var}_mul')

        Examples:
            cc = FromDatasetCalculator(...)
            cc_scaled1 = cc.with_scaling({'n': 1e-2})  # simple: scale densities by 1e-2.
            # testing multiple scalings at once: add a dimension!
            cc_scaled2 = cc.with_scaling({'n': pc.xr1d([1, 1e-2, 1e-4], name='n_mul')})
            # shorthand for above; internally uses pc.xr1d, name='{var}_mul'.
            cc_scaled3 = cc.with_scaling({'n': [1, 1e-2, 1e-4]})  # equivalent to cc_scaled2.
        '''
        scales = scaling_dict.copy()
        for var, value in scaling_dict.items():
            if is_iterable(value) and not isinstance(value, xr.DataArray):
                scales[var] = xr1d(np.array(value), name=f'{var}_mul')
        scaled_ds = self.ds.assign({v: scales[v] * self.ds[v] for v in scales})
        return type(self)(scaled_ds, **kw_init)


''' --------------------- Inferring Dimensions From Dataset --------------------- '''

class InfersSnapsFromDataset():
    '''adds self.init_snaps() which infers snaps from self.ds['snap'] if available.
        (if ds['snap'] contents are not Snap objects, infer 't' from ds['t'] too if available.)
    '''
    snap_list_cls = SnapList

    def __init__(self, ds, **kw_super):
        super().__init__(ds, **kw_super)
        self.init_snaps()

    def init_snaps(self):
        '''initialize snaps from self.ds['snap'] if available.
        (Otherwise, self will have the default self.snap=None, self.snaps=None.)
        '''
        if 'snap' in self.ds:
            self.snaps = self.snap_list_cls.from_array(self.ds['snap'])
            self.snap = self.snaps[0] if self.ds['snap'].ndim == 0 else None

    def load_direct(self, var, *args, **kw):
        '''load var directly from self.ds if possible. slice snap dimension based on self.snap.'''
        result = super().load_direct(var, *args, **kw)
        if (self.snap is not self.snaps) and ('snap' in result.coords):
            result = xarray_sel(result, snap=self.snap)
        return result


class InfersFluidsFromDataset():
    '''adds self.init_fluids() which infers fluids and jfluids from self.ds['fluid'] and 'jfluid' if available.'''
    fluid_list_cls = FluidList
    jfluid_list_cls = alias('fluid_list_cls')

    def __init__(self, ds, **kw_super):
        super().__init__(ds, **kw_super)
        self.init_fluids()

    def init_fluids(self):
        '''initialize fluids and jfluids from self.ds['fluid'] and ds['jfluid'] if available.
        (Otherwise, self will have the default: self.fluid, fluids, jfluid, and jfluids=None.)
        '''
        if 'fluid' in self.ds:
            self.fluids = self.fluid_list_cls.from_array(self.ds['fluid'])
            self.fluid = self.fluids[0] if self.ds['fluid'].ndim == 0 else None
        if 'jfluid' in self.ds:
            self.jfluids = self.jfluid_list_cls.from_array(self.ds['jfluid'])
            self.jfluid = self.jfluids[0] if self.ds['jfluid'].ndim == 0 else None

    def load_direct(self, var, *args, **kw):
        '''load var directly from self.ds if possible. slice fluid & jfluid dims based on self.fluid & jfluids.'''
        result = super().load_direct(var, *args, **kw)
        if (self.fluid is not self.fluids) and ('fluid' in result.coords):
            result = xarray_sel(result, fluid=self.fluid)
        if (self.jfluid is not self.jfluids) and ('jfluid' in result.coords):
            result = xarray_sel(result, jfluid=self.jfluid)
        return result


class InfersComponentsFromDataset():
    '''adds self.init_components() which infers components from self.ds['component'] if available.'''
    component_list_cls = ComponentList

    def __init__(self, ds, **kw_super):
        super().__init__(ds, **kw_super)
        self.init_components()

    def init_components(self):
        '''initialize components from self.ds['component'] if available.
        (Otherwise, self will have the default self.component=None, self.components=None.)
        '''
        if 'component' in self.ds:
            self.components = self.component_list_cls.from_array(self.ds['component'])
            self.component = self.components[0] if self.ds['component'].ndim == 0 else None

    def load_direct(self, var, *args, **kw):
        '''load var directly from self.ds if possible. slice component dimension based on self.component.'''
        result = super().load_direct(var, *args, **kw)
        if (self.component is not self.components) and ('component' in result.coords):
            result = xarray_sel(result, component=self.component)
        return result


''' --------------------- Dimensional From Dataset Calculators --------------------- '''

class ComponentHavingFromDatasetCalculator(InfersComponentsFromDataset,
                                           DimensionlessFromDatasetCalculator,
                                           ComponentHavingPlasmaCalculator):
    '''DimensionlessFromDatasetCalculator but with vector arithmetic and derivatives.
    Infers components from dataset if possible.

    ds: xarray.Dataset
        self(var) looks in self.ds when can't find another way to compute var.
        For base vars, always checks ds first.
        (base vars): ds, q, gamma, E, B, m, n, u, T, nusj,
                     m_neutral, n_neutral, u_neutral, T_neutral.
        For simple derived vars, check ds first; compute from base vars if missing.
        (simple derived vars): r, p, P, Tjoule, e, nusn, nq, Jf, J, E_un0.
    '''
    BASES_CHECK = ['ds', 'q', 'gamma', 'E', 'B', 'm', 'n', 'u', 'T', 'nusj',
                    'm_neutral', 'n_neutral', 'u_neutral', 'T_neutral',
                    'nusn', 'E_un0',]
    # all other functionality inherited by super().


class FromDatasetCalculator(InfersSnapsFromDataset,
                            ComponentHavingFromDatasetCalculator,
                            PlasmaCalculator):
    '''PlasmaCalculator taking a Dataset as input. Knows about snaps and components dimensions.
    Infers snaps and components from dataset if possible.

    ds: xarray.Dataset
        self(var) looks in self.ds when can't find another way to compute var.
        For base vars, always checks ds first.
        (base vars): ds, q, gamma, E, B, m, n, u, T, nusj,
                         m_neutral, n_neutral, u_neutral, T_neutral.
        For simple derived vars, check ds first; compute from base vars if missing.
        (simple derived vars): r, p, P, Tjoule, e, nusn, nq, Jf, J, E_un0.
    '''
    pass


class MultifluidFromDatasetCalculator(InfersFluidsFromDataset,
                                      FromDatasetCalculator,
                                      MultifluidPlasmaCalculator):
    '''PlasmaCalculator taking a Dataset as input. Knows about snaps, components, and fluids.
    Infers snaps, components, fluids, and jfluids from dataset if possible.

    ds: xarray.Dataset
        self(var) looks in self.ds when can't find another way to compute var.
        For base vars, always checks ds first.
        (base vars): ds, q, gamma, E, B, m, n, u, T, nusj,
                         m_neutral, n_neutral, u_neutral, T_neutral.
        For simple derived vars, check ds first; compute from base vars if missing.
        (simple derived vars): r, p, P, Tjoule, e, nusn, nq, Jf, J, E_un0.
    '''
    pass


class VectorlessFromDatasetCalculator(InfersSnapsFromDataset,
                                      DimensionlessFromDatasetCalculator,
                                      VectorlessPlasmaCalculator):
    '''FromDatasetCalculator but without vector components/arithmetic/derivatives.
    Has all other features of FromDatasetCalculator.
    Infers snaps from dataset if possible.

    ds: xarray.Dataset
        self(var) looks in self.ds when can't find another way to compute var.
        For base vars, always checks ds first.
        (base vars): ds, q, gamma, mod_E, mod_B, E_perpmod_B, E_un0_perpmod_B,
                     m, n, mod_u, T, nusj,
                     m_neutral, n_neutral, mod_u_neutral, T_neutral.
        For simple derived vars, check ds first; compute from base vars if missing.
        (simple derived vars): r, p, P, Tjoule, e, nusn, nq, Jf, J, E_un0.
    '''
    pass


class VectorlessMultifluidFromDatasetCalculator(InfersFluidsFromDataset,
                                                VectorlessFromDatasetCalculator,
                                                VectorlessMultifluidPlasmaCalculator):
    '''MultifluidFromDatasetCalculator but without vector components/arithmetic/derivatives.
    Has all other features of MultifluidFromDatasetCalculator.
    Infers snaps, fluids, and jfluids from dataset if possible.

    ds: xarray.Dataset
        self(var) looks in self.ds when can't find another way to compute var.
        For base vars, always checks ds first.
        (base vars): ds, q, gamma, mod_E, mod_B, E_perpmod_B, E_un0_perpmod_B,
                     m, n, mod_u, T, nusj,
                     m_neutral, n_neutral, mod_u_neutral, T_neutral.
        For simple derived vars, check ds first; compute from base vars if missing.
        (simple derived vars): r, p, P, Tjoule, e, nusn, nq, Jf, J, E_un0.
    '''
    pass



"""
File purpose: manage plot settings & kwargs, to make the other code more readable

Note that the code here doesn't enforce any particular values or relationships for kwargs,
it just makes it easier to manage kwargs & docs for all these settings.

[TODO] add list of kwargs for mpl funcs as options in format_docstring.
E.g. want to say "additional kwargs go to FuncAnimation.save, particularly: {list of kwargs}"
"""

import inspect
import textwrap

import matplotlib.pyplot as plt

from ..errors import InputError, InputConflictError, PlotSettingsError
from ..tools import (
    UNSET, NO_VALUE, RESULT_MISSING,
    format_docstring,
)
from ..defaults import DEFAULTS


# # # DEFAULT SETTINGS FOR PLOTS # # #
DEFAULT_PLOT_SETTINGS = {
    # key: (default, type, docstring, (key in DEFAULTS; optional))
    # will all be converted to PlotSetting instances.

    # key: [subkeys, type, docstring, (key in DEFAULTS; optional)]
    # will all be converted to SharedPlotSetting instances.

    # # # GENERIC # # #
    'dpi': (UNSET, "UNSET, None, or number",
        """dots per inch... used by various matplotlib functions.""",
        ),
    # # # PLOT LIMITS / MARGINS # # #
    'kw_ax_lim': [
        ('xlim', 'ylim'),
            (None, "None or 2-tuple of None or number",
            """plt.xlim / plt.ylim for each plot. None --> don't adjust plot limits."""),
        ],
    'kw_ax_margin': [
        ('xmargin', 'ymargin', 'margin'),
            (None, "None or number (greater than -0.5, probably close to 0.05)",
            r"""margin to use for x/y axis, as a fraction of the data interval for that axis.
            None --> use matplotlib defaults
                    (e.g., plt.rcParams["axes.xmargin"] or ["axes.ymargin"], or 0 if using imshow)
            positive number --> pad around the data region, with this much whitespace.
                    E.g. 0.05 means adding 5% whitespace on each side.
                    Use this to zoom out.
            negative number --> remove this much of the outer parts of the data region.
                    E.g. -0.2 means removing 20% space from each side.
                    Use this to zoom in.
            For line plots, if also using `robust`, ymargin will be applied to the robust y lims.
            (margin-related params share the same docstring, but refer to:
                'xmargin': x-axis, 'ymargin': y-axis, 'margin': x and/or y-axis.)""",),
        ],
    'robust': (UNSET, "UNSET, bool, or number between 0 and 50",
        """use np.percentile when determining vmin/vmax, if robust.
        For imshow/image plots, this refers to colorbar lims; for line plots, this refers to y lims.
        UNSET --> use {default}.
        False --> just use min and max of values, don't use percentile.
        True --> use {defaults_dict_True}
        number --> use np.percentile with this percentile for vmin (and 100 - this percentile for vmax).""",
        ('DEFAULTS.PLOT.ROBUST', {True: 'DEFAULTS.PLOT.ROBUST_PERCENTILE'}),
        ),
    'ybounds': (None, "None, False, or 2-tuple/iterable of [max_ymin, min_ymax]",
        """tells the (maximum ymin, minimum ymax) when determining ylims.
            None --> use the current ylims if current_axes_has_data(), else (None, None).
            False --> equivalent to (None, None), i.e. ignore this setting.
        E.g., overlaying multiple plots; plot 1 from -10 to 10, plot 2 from -5 to 15;
        by default would use plot 2 ybounds=(-10, 10) --> final ylims of (-10, 15).
        If provided plot 2 ybounds=(-7, 12), would instead have ylims of (-7, 15).""",
        ),
    # # # PLOT TICKS # # #
    'min_n_ticks': (UNSET, "UNSET, None, int, or 2-tuple of ints",
        """minimum number of ticks to use on plot axes, e.g. x & y axes.
        UNSET --> use {default}.
        None --> use matplotlib default (probably 2).
        int --> use this number of ticks.
        tuple --> provides (min for x, min for y).""",
        'DEFAULTS.PLOT.MIN_N_TICKS',
        ),
    'min_n_ticks_fail_ok': (UNSET, "bool",
        """whether it is okay for set_min_n_ticks() to fail silently in general,
        by default, when called from PlasmaCalcs plotting codes.
        UNSET --> use {default}.""",
        'DEFAULTS.PLOT.MIN_N_TICKS_FAIL_OK',
        ),
    'min_n_ticks_cbar': (UNSET, "UNSET, None, int, or 2-tuple of ints",
        """minimum number of ticks to use on colorbars.
        UNSET --> use {default}.
        None --> use matplotlib default (probably 2).
        int --> use this number of ticks.
        tuple --> provides (min for horizontal cbars, min for vertical cbars).""",
        'DEFAULTS.PLOT.MIN_N_TICKS_CBAR',
        ),
    'min_n_ticks_cbar_fail_ok': (UNSET, "bool",
        """whether it is okay for set_min_n_ticks() to fail silently for colorbars,
        by default, when called from PlasmaCalcs plotting codes.
        True enables to specify x and y minimums simultaneously,
            while only applying x mins to horizontal cbars, and y mins to vertical cbars.
        UNSET --> use {default}.""",
        'DEFAULTS.PLOT.MIN_N_TICKS_CBAR_FAIL_OK',
        ),
    # # # SUBPLOTS # # #
    'nrows': (1, "int",
        """number of rows in the subplots""",
        ),
    'ncols': (1, "int",
        """number of columns in the subplots""",
        ),
    'axsize': (UNSET, "UNSET, number, or (width, height) in inches",
        """size of a single subplot, in inches.
        UNSET --> use {default}.
        number --> use width = height = axsize.
        mutually exclusive with figsize, cannot provide both.""",
        'DEFAULTS.PLOT.SUBPLOTS_AXSIZE',
        ),
    'figsize': (UNSET, "UNSET, None, or (width, height) in inches",
        """if provided (even if None) pass to matplotlib.
        otherwise, for subplots, use axsize to determine figsize.""",
        ),
    'squeeze': (False, "False",
        """To avoid ambiguity, squeeze=True is not allowed; Subplots.axs will always be 2D.""",
        ),
    'kw_subplots_adjust': [
        ('hspace', 'wspace', 'bottom', 'top', 'left', 'right'),
            (None, "None or number (probably between 0 and 1)",
            """corresponding value that will be used during plt.subplots_adjust(...).
            If any region of a saved plot is getting cut off, try adjusting these values.
            None --> use rcParams["figure.subplot.*"]."""),
        ],
    'kw_share_ax': [
        ('sharex', 'sharey'),
            (None, "None, bool, or str ('all', 'row', 'col')",
            """whether & how to share x & y axes.
            None --> use False, by default, due to a formatting issue with sharex/sharey and make_cax.
            if provided, also set sharexlike and shareylike appropriately."""),
        ],
    'kw_share_axlike': [
        ('sharexlike', 'shareylike'),
            (UNSET, "UNSET or bool",
            """whether ticks and plot labels behave like there is a shared x, y axis.
            True --> remove tick labels for all but the bottom/left subplots,
                and Subplots.xlabel and .ylabel will only affect bottom/left subplots, by default.
            False --> retain all tick labels and all axes labels.
            UNSET --> use sharex/sharey if provided, else infer based on existing labels & ticks."""),
        ],
    'max_nrows_ncols': (UNSET, "UNSET, None, or int",
        """maximum number of rows or cols in subplots grid before crashing.
        UNSET --> use {default}
        None --> no maximum. Use maxlen=None to turn off the maximum if making a very large plot.""",
        'DEFAULTS.PLOT.SUBPLOTS_MAX_NROWS_NCOLS',
        ),
    'grid': (UNSET, "UNSET, bool, or dict",
        """whether to ax.grid() for each axes.
        UNSET --> use rcParams["axes.grid"].
        True --> ax.grid(True) for all axes.
        False --> ax.grid(False) for all axes.
        dict --> ax.grid(**grid) for all axes.""",
        ),
    # # # MOVIE (GENERIC) # # #
    'frames': (UNSET, "UNSET, None, int, iterable, or slice",
        """passed to FuncAnimation. Tells number of frames or which frames to plot.
        If UNSET, use value from self.plot_settings if possible else getattr(self, 'frames', None).
        if slice, use range(self.get_nframes())[frames], crashing if self doesn't have 'get_nframes'.""",
        ),
    'fps': (UNSET, "UNSET, None, or number",
        """frames per second.
        UNSET --> use {default}.
                (Or use value from self.plot_settings, if provided.)
        None --> use matplotlib defaults.""",
        'DEFAULTS.PLOT.FPS',
        ),
    'blit': (UNSET, "UNSET, None, or bool",
        """whether to use blitting.
        UNSET --> use {default}.
                (Or use value from self.plot_settings, if provided.)
        if None, use matplotlib defaults.""",
        'DEFAULTS.PLOT.BLIT',
        ),
    'progress_callback': (UNSET, "UNSET, None, or callable",
        """called when rendering each frame of the animation, as:
            progress_callback(current frame number, total number of frames).
        UNSET --> default to lambda i, n: updater.print('saving frame '+str(i)+' of '+str(n)),
                    where updater = ProgressUpdater(print_freq=1 if verbose else -1).
                    (Or use value from self.plot_settings, if provided)
        None --> don't call anything, regardless of self.verbose.""",
        ),
    # # # MOVIE IMAGE # # #
    'image_mode': ('pcolormesh', "str ('imshow' or 'pcolormesh')",
        """tells whether this image will be pcolormesh or imshow.""",
        None,  # no value from DEFAULTS
        ('pcolormesh', 'imshow'),  # apply validation immediately when setting; value must be one of these.
        ),
    'init_plot': (True, "bool",
        """whether to call self.init_plot() immediately, during __init__.
        (as a PlotSetting: tells the value of init_plot passed during __init__.)
        False --> still must call self.init_plot() before using self(...) or self.updater(...)
            (end-user will probably never use init_plot=False; it's mostly for internal code.)
            (might use False if creating MovieImage before defining the associated Axes.)""",
        ),
    'init_plot_frame': (0, "int",
        """default frame to use when calling self.init_plot()""",
        ),
    # # # XARRAY IMAGE # # #
    'add_colorbar': (UNSET, "UNSET or bool",
        """if provided, default for add_colorbar when making xarray plots.""",
        ),
    'add_labels': (True, "bool",
        """whether to add labels to xarray plots.""",
        ),
    # # # PATCHES # # #
    'transform': (UNSET, "UNSET, 'data', 'axes', 2-tuple of ('data' or 'axes'), or Transform object",
        """indicate coordinate system to use for x and y inputs.
        single string --> x and y both in this coordinate system.
        tuple --> xy[0] tells x system; xy[1] tells y system.
        'data' coords means input values match data values.
        'axes' coords means input values correspond to distance across axis:
            for x: left=0, right=1.
            for y: bottom=0, top=1.""",
        ),
    # # # UPDATABLE TEXT / TITLE / SUPTITLE # # #
    'text_kw': (UNSET, "UNSET, None, or dict",
        """any additional kwargs for text, e.g. text_kw=dict(fontweight=bold).
        Applied to titles too, but individual kw can be overridden by title settings.
        [TODO] ability to enter unambiguous text_kw directly, e.g. fontweight.""",
        ),
    'title': (UNSET, "UNSET, None, or str",
        """title for a single axes/subplot. For xarrays, will title.format(**xarray_nondim_coords(array)).
        (Note: plot_settings['title'] should always be the 'base' title, before title.format(...))
        UNSET --> use array_at_current_frame._title_for_slice() if XarrayImage (or other single-array plot).
                  use XarraySubplotTitlesInferer.infer_title() if XarraySubplots (or other multi-array plot).
        None --> do not add a title.""",
        ),
    'title_font': (UNSET, "UNSET, None, or str",
        """font for title, e.g. 'serif', 'sans-serif', or 'monospace'
        UNSET --> use {default}.
        None --> use matplotlib default.""",
        'DEFAULTS.PLOT.MOVIE_TITLE_FONT',
        ),
    'title_y': (UNSET, "UNSET, None, or number",
        """y position for title, in axes coordinates.""",
        ),
    'title_kw': (UNSET, "UNSET, or dict",
        """any additional kwargs for plt.title.""",
        ),
    'subplot_title_width': (UNSET, "UNSET, None, or number",
        """suggested width [number of characters] for subplot titles;
        default title routines might make multiline title if title would be longer than this.
        UNSET --> use {default}.
        None --> no maximum width.""",
        'DEFAULTS.PLOT.SUBPLOT_TITLE_WIDTH',
        ),
    'rtitle': (UNSET, "UNSET or str",
        """rightmost-column 'title' to put only on righthand side of subplots on the rightmost column.
        For xarrays, will rtitle.format(**xarray_nondim_coords(array)).
        Note: rtitles created via PlasmaCalcs' plot_note(), which uses plt.annotate, not plt.ylabel().
        UNSET --> do not add rtitle.""",
        ),
    'rtitle_kw': (UNSET, "UNSET, or dict",
        """any additional kwargs for plot_note() when making rtitle.
        (note: includes font=plot_settings['title_font'] unless `fontfamily` specified in rtitle_kw.)
        (note: to specify location, use `loc` or `xy` in axes coords or as str; see plot_note() for details.)
        UNSET --> use {default}.""",
        'DEFAULTS.PLOT.RTITLE_KW',
        ),
    'ttitle': (UNSET, "UNSET or str",
        """topmost-row title to put only on subplots in the top row.
        For xarrays, will ttitle.format(**xarray_nondim_coords(array)).
        Mutually exclusive with providing `title`, and overrides default titles if provided.
        UNSET --> do not add ttitle.""",
        ),
    'ttitle_kw': (UNSET, "UNSET, or dict",
        """any additional kwargs for plt.title when making ttitle.
        (note: includes font=plot_settings['title_font'] unless `fontfamily` specified in ttitle_kw.)
        UNSET --> use {default}.
        None --> use no title.""",
        'DEFAULTS.PLOT.TTITLE_KW',
        ),
    'suptitle': (UNSET, "UNSET, None, or str",
        """suptitle for a single axes/subplot. For xarrays, will suptitle.format(**xarray_nondim_coords(array)).
        (Note: plot_settings['suptitle'] should always be the 'base' suptitle, before suptitle.format(...))
        UNSET --> use self.default_suptitle. For XarraySubplots this includes information about t_plot_dim.
        None --> use no suptitle.
        (Note - self.plot_settings['suptitle'] will always be the 'base' suptitle, before applying suptitle_format.)""",
        ),
    'suptitle_font': (UNSET, "UNSET, None, or str",
        """font for suptitle, e.g. 'serif', 'sans-serif', or 'monospace'
        UNSET --> use {default}.
        None --> use matplotlib default.""",
        'DEFAULTS.PLOT.MOVIE_TITLE_FONT',
        ),
    'suptitle_y': (UNSET, "UNSET, None, or number",
        """y position for suptitle, in figure coordinates.""",
        ),
    'suptitle_kw': (UNSET, "UNSET, or dict",
        """any additional kwargs for plt.suptitle.""",
        ),
    'suptitle_width': (UNSET, "UNSET, None, or number",
        """suggested width [number of characters] for suptitle;
        default routines might make multiline suptitle to avoid going longer than this.
        UNSET --> use {default}.
        None --> no maximum width.""",
        'DEFAULTS.PLOT.SUPTITLE_WIDTH',
        ),
    # # # IMAGE SUBPLOTS # # #
    'add_colorbars': (True, "UNSET, bool, or str",
        """whether to add colorbars during init_plot, for ImageSubplots.
        str --> use self.colorbars(mode=add_colorbars). E.g. 'auto', 'all', 'row'.""",
        ),
    'aspect': (UNSET, "UNSET, None, str, or number",
        """aspect ratio for each Axes, by default.
        UNSET --> use {default}.
        None --> use matplotlib defaults.
        str --> use 'equal', or 'auto'. Note that 'equal' is equivalent to using aspect=1.
        number --> height scaling / width scaling.
                E.g. aspect=2 --> 1 data unit of height is twice long as 1 data unit of width.""",
        'DEFAULTS.PLOT.ASPECT',
        ),
    'layout': (UNSET, "UNSET, None, or str",
        """layout for subplots, by default.
        Suggestion: use layout='compressed', make_cax='mpl', and tweak wspace & hspace,
            OR use layout='none', make_cax='pc', and tweak suptitle_y, left, top, and bottom.
        UNSET --> use {default}.
        None --> use matplotlib defaults.
        str --> should be 'constrained', 'compressed', 'tight', or 'none'.""",
        'DEFAULTS.PLOT.LAYOUT',
        ),
    'share_vlims': (False, "bool or str ('all', 'row', 'col')",
        """whether to share vmin/vmax across ImageSubplots.
        True --> use 'all'
        'all' --> share vlims across all image subplots.
        'row' --> share vlims across each row of image subplots.
        'col' --> share vlims across each column of image subplots.""",
        ),
    # # # COLORBARS # # #
    'cax_mode': (UNSET, "UNSET, 'mpl', 'pc'",
        """tells how to make a new axis for a colorbar if one was not provided.
        UNSET --> use {default}.
        'mpl' --> use matplotlib logic, the same exact logic as in plt.colorbar.
        'pc' --> use PlasmaCalcs logic; it looks better when using layout='none',
                however 'pc' logic doesn't play nicely with other layout options yet.""",
        'DEFAULTS.PLOT.CAX_MODE',
        ),
    'colorbar_kw': (UNSET, "unset or dict",
        """any additional kwargs for plt.colorbar.""",
        ),
    # # # 3D FACEPLOTS # # #
    'faceplot_view_angle': (UNSET, "UNSET, None, or 3-tuple of numbers",
        """viewing angle for 3D faceplots, as (elevation, azimuth, roll).
        UNSET --> use {default}.
        None --> use matplotlib defaults.""",
        'DEFAULTS.PLOT.FACEPLOT_VIEW_ANGLE',
        ),
    'faceplot_edge_kwargs': (UNSET, "UNSET, None, or dict",
        """kwargs for edge lines in 3D faceplots.
        UNSET --> use {default}.
        None --> don't plot edge lines.
        empty dict --> use matplotlib defaults.""",
        'DEFAULTS.PLOT.FACEPLOT_EDGE_KWARGS',
        ),
    'faceplot_axes_zoom': (UNSET, "UNSET or number>0",
        """zoom for faceplot axis. matplotlib default is zoom=1.
        UNSET --> use {default}.""",
        'DEFAULTS.PLOT.FACEPLOT_AXES_ZOOM',
        ),
    'aspect3d': (UNSET, "UNSET, None, str, 3-tuple of numbers, or 4-tuple of numbers",
        """aspect ratio for 3D plots.
        UNSET --> use {default}.
        str --> 'auto' or 'equal'
        tuple of 3 numbers --> (x aspect, y aspect, z aspect)
        tuple of 4 numbers --> (1, x multiplier, y multiplier, z multiplier);
                multiplier multiplies aspect determined by data lengths.""",
        'DEFAULTS.PLOT.ASPECT3D',
        ),
    'proj_type': (UNSET, "UNSET, 'persp', or 'ortho'",
        """projection type for 3D plots. For details see mpl_toolkits.mplot3d.axes3d.Axes3D.
        UNSET --> use {default}.""",
        'DEFAULTS.PLOT.PROJ_TYPE',
        ),
    # # # CONTOUR # # #
    'colorbar_linewidth': (UNSET, "UNSET, None, int, list-like, or 2-tuple of None/int.",
        """linewidth for lines in contour colorbar.
        UNSET --> use {default}.
        None --> use same width as contour lines.
        int or list-like --> use this as the linewidth
                (as per matplotlib.collections.LineCollection.set(linewidth=...))
        tuple --> defines (min, max) linewidths; use None for no bound.
                E.g. (4, None) says "for thinner lines use 4; others same as contour lines".""",
        'DEFAULTS.PLOT.COLORBAR_LINEWIDTH',
        ),
    'colorbar_linestyle': (UNSET, "UNSET, None, or str",
        """linestyle for lines in contour colorbar.
        UNSET --> use {default}.
        None --> use same style as contour lines.""",
        'DEFAULTS.PLOT.COLORBAR_LINESTYLE',
        ),
    'colorbar_lines_set': (UNSET, "UNSET, or dict",
        """any additional attrs to set for contour colorbar lines,
        via cbar.ax.get_children()[1].set(...)""",
        ),
    'label': (UNSET, "UNSET or str",
        """label for this plot element, to be included in a legend.
        For xarrays, will label.format(**xarray_nondim_coords(array)).""",
        ),
    'legend_handle_kw': (UNSET, "UNSET or dict",
        """any additional kwargs to use when creating legend handles.""",
        ),
    'contourf': (False, "bool",
        """whether to use filled contours (i.e., plt.contourf) or not (i.e., plt.contour).""",
        ),
}

# # # ALL KWARGS TO PASS DIRECTLY TO MATPLOTLIB FUNCTIONS # # #
MPL_KWARGS = {
    # these should be the kwargs which are passed to matplotlib functions.
    # 'here' means the kwarg is used by that matplotlib function directly.
    # 'from' means "also accept any kwargs from these other matplotlib function(s)"
    # other keys mean "pass these kwargs as this key=dict(other kwargs)".
    #    (e.g. plt.subplots > subplot_kw > from mpl.figure.Figure.add_subplot means
    #    plt.subplots(..., subplot_kw=dict(**(kwargs from mpl.figure.Figure.add_subplot)))

    # ... actually, this is compatible with any function, not just matplotlib functions.
    # [TODO] clarify this, somehow.

    'plt.figure': {
        'here': ('figsize', 'dpi', 'facecolor', 'edgecolor', 'FigureClass',
                  'clear', 'layout', 'zorder'),
        },
    'fig.subplots': {
        'here': ('width_ratios', 'height_ratios', 'squeeze'),
        'subplot_kw': {
            'from': ['mpl.figure.Figure.add_subplot'],
            },
        },
    'plt.subplots': {
        'from': ['plt.figure', 'fig.subplots'],
        },
    'plt.subplots_adjust': {
        'here': ('left', 'bottom', 'right', 'top', 'wspace', 'hspace'),
        },
    'mpl.figure.Figure.add_subplot': {
        'here': ('projection', 'polar', 'axes_class',
                 'aspect'),
        },
    'mpl.animation.FuncAnimation': {
        'here': ('blit', 'frames', 'save_count', 'repeat', 'cache_frame_data',
                 'repeat_delay'),
        },
    'mpl.animation.FuncAnimation.save': {
        'here': ('codec', 'bitrate', 'metadata', 'savefig_kwargs', 'dpi'),
        },
    'xarray.DataArray.plot': {
        'here': ('xscale', 'yscale', 'xlim', 'ylim', 'xincrease', 'yincrease',
                 #'aspect',  # <-- don't include aspect here since it is finicky (requires "size" too)
                 ),
        },
    'xarray.DataArray.plot[image]': {   # for arr.plot.imshow & arr.plot.pcolormesh
        'here': ('cmap', 'vmin', 'vmax', 'add_colorbar', 'add_labels',
                'robust',
                'subplot_kws',
                ),
        'from': ['xarray.DataArray.plot'],
        },
    'xarray.DataArray.plot.contour': {
        'here': (),
        'from': ['xarray.DataArray.plot[image]', 'plt.contour'],
        },
    'xarray.DataArray.plot.contourf': {
        'here': (),
        'from': ['xarray.DataArray.plot[image]', 'plt.contourf'],
        },
    'xarray.DataArray.plot.line': {
        'here': (),
        'from': ['xarray.DataArray.plot', 'plt.plot'],
        },
    'xarray.DataArray.plot.scatter': {
        'here': ('add_labels',),
        'from': ['xarray.DataArray.plot', 'plt.scatter'],
        },
    'plt.colorbar': {
        'here': ('mappable', 'cax', 'use_gridspec', 'location', 'orientation',
                'fraction', 'shrink', 'aspect', 'pad', 'anchor', 'panchor', 'extend',
                'extendfrac', 'extendrect', 'spacing', 'ticks', 'format', 'drawedges',
                'label', 'boundaries', 'values',
                ),
        },
    'mpl.lines.Line2D': {
        'here': ('alpha', 'animated', 'antialiased', 'clip_box', 'clip_on', 'clip_path',
                'color', 'dash_capstyle', 'dash_joinstyle', 'dashes', 'drawstyle',
                'fillstyle', 'label', 'linestyle', 'ls', 'linewidth', 'lw', 'marker',
                'markeredgecolor', 'mec', 'markeredgewidth', 'mew', 'markerfacecolor', 'mfc',
                'markersize', 'ms', 'markevery', 'solid_capstyle', 'solid_joinstyle',
                'visible', 'zorder',
                ),
    },
    'plt.plot': {
        'here': ('scalex', 'scaley'),
        'from': ['mpl.lines.Line2D'],
        },
    'plt.errorbar': {
        'here': ('ecolor', 'elinewidth', 'capsize', 'capthick', 'barsabove',
                'lolims', 'uplims', 'xlolims', 'xuplims',
                'errorevery',
                ),
        'from': ['mpl.lines.Line2D'],
        },
    'mpl.collections.Collection': {
        'here': ('edgecolors', 'facecolors', 'linewidths', 'linestyles', 'capstyle',
                 'joinstyle', 'offsets', 'cmap', 'norm', 'hatch', 'pickradius', 'zorder',
                 # implicit, available via set_{key}(val):
                 'alpha', 'animated', 'antialiased', 'clip_box', 'clip_on', 'clip_path',
                 'color', 'edgecolor', 'facecolor', 'hatch_linewidth', 'in_layout',
                 'joinstyle', 'label', 'linestyle', 'linewidth', 'mouseover',
                 'path_effects', 'paths', 'picker', 'visible',
                 ),
    },
    'plt.fill_between': {
        'here': ('step',),
        'from': ['mpl.collections.Collection'],
    },
    'plt.contour': {
        'here': ('colors', 'alpha', 'cmap', 'vmin', 'vmax', 'levels',
                 'linestyles', 'linewidths',
                 ),
        },
    'plt.contourf': {
        'here': ('colors', 'alpha', 'cmap', 'vmin', 'vmax', 'levels',
                 'hatches',
                 ),
        },
    'plt.legend': {
        'here': ('loc', 'handlelength', 'bbox_to_anchor', 'ncols',
                 ),  # [TODO] more options for plt.legend.
        },
    'plt.imshow': {
        'here': ('cmap', 'vmin', 'vmax', 'norm', 'alpha',
                 'aspect', 'interpolation', 'origin', 'extent',
                 ),
        },
    'plt.pcolormesh': {
        'here': ('cmap', 'vmin', 'vmax', 'norm', 'alpha',
                 'edgecolors', 'shading', 'snap', 'rasterized',
                 ),
        },
    'plt.scatter': {
        'here': ('s', 'c', 'marker', 'cmap', 'norm', 'vmin', 'vmax', 'alpha',
                 'linewidths', 'edgecolors', 'colorizer', 'plotnonfinite',
                 'facecolors', 'label', 'visible', 'color',
                 ),
        },
    'ax.set': {
        'here': ('xlabel', 'ylabel', 'zlabel', 'xlim', 'ylim', 'zlim',
                 'xticks', 'yticks', 'zticks', 'xticklabels', 'yticklabels', 'zticklabels',
                 ),
        },
    'mpl_toolkits.mplot3d.axes3d.Axes3D': {
        'here': ('elev', 'azim', 'roll', 'proj_type', 'focal_length', 'shareview',
                 ),
        },
    'mpl.patches.Patch': {
        'here': ('alpha', 'animated', 'antialiased', 'capstyle', 'clip_box', 'clip_on', 'clip_path',
                 'color', 'edgecolor', 'ec', 'facecolor', 'fc', 'fill', 'hatch', 'hatch_linewidth',
                 'in_layout', 'joinstyle', 'label', 'linestyle', 'ls', 'linewidth', 'lw',
                 'mouseover', 'path_effects', 'picker', 'rasterized', 'sketch_params', 'snap',
                 'visible', 'zorder',
                 ),
        },
    'mpl.patches.Rectangle': {
        'here': ('width', 'height', 'angle', 'rotation_point'),
        'from': ['mpl.patches.Patch'],
        },
    # technically the rest of these aren't matplotlib functions but... still useful to have.
    'pc.TextPlotElement': {
        'here': ('text_kw',),
        },
    'pc.XarrayImagePlotElement': {
        'here': ('image_mode', 'aspect', 'xlabel', 'ylabel', 'grid', 'polar',
                 'min_n_ticks', 'min_n_ticks_cbar',
                 ),
        'from': ['xarray.DataArray.plot[image]', 'plt.figure'],
        },
    'pc.XarraySubplotTitlesInferer': {
        'here': ('subplot_title_width', 'suptitle_width'),
        },
    'pc.XarrayContourPlotElement': {
        'here': ('add_colorbar', 'add_labels', 'contourf',
                 'label', 'legend_handle_kw',
                 'colorbar_linewidth', 'colorbar_linestyle', 'colorbar_lines_set',
                 'aspect',
                 ),
        'from': ['xarray.DataArray.plot.contour', 'xarray.DataArray.plot.contourf'],
        },
    'pc.FaceplotPlotElement': {  
        'here': ('faceplot_view_angle', 'faceplot_edge_kwargs', 'faceplot_axes_zoom',
                 'aspect3d',
                 'add_colorbar', 'colorbar_kw',
                 ),
        'from': ['ax.set', 'plt.contourf'],
        },
    'pc.XarrayLinePlotElement': {
        'here': ('add_labels', 'label', 'aspect',
                 'robust', 'ymargin',
                 ),
        'from': ['xarray.DataArray.plot.line'],
        },
    'pc.ScatterPlotElement': {
        'here': (
                 # it always bugs me that these aliases aren't options by default in plt.scatter...:
                 'markersize',  # alias to 's'
                 'color',  # alias to 'c'
                 'facecolor',  # alias to 'facecolors'
                 'edgecolor',  # alias to 'edgecolors'
                 'linewidth',  # alias to 'linewidths'
                 ),
        'from': ['xarray.DataArray.plot.scatter'],
    },
    'pc.PatchPlotElement': {
        'here': ('transform',),
        'from': ['mpl.patches.Patch'],
        },
    'pc.RectanglePatchPlotElement': {
        'from': ['pc.PatchPlotElement', 'mpl.patches.Rectangle'],
        },
}


''' --------------------- PlotSetting and PlotSettings classes --------------------- '''

class PlotSetting():
    '''a single plot setting, with default value.
    key: key for this setting.
    default: default value for this setting
    typsetr: string describing the type for this setting
    doc: docstring for this setting (not including the typestr) (will be hit by inspect.cleandoc)
    defaults_keystring: None, string of the form 'DEFAULTS.<key>', or tuple of (defaults_keystring, dict)
        if string, self.get(provided=UNSET) will get value from DEFAULTS.
        if tuple, first element must be str, and first behaves as if just a str,
            but then based on result of self.get, might use dict as follows:
            if val=self.get(provided=UNSET) in dict, use value from DEFAULTS implied by dict[val].
            E.g. ('DEFAULTS.PLOT.ROBUST', {True: 'DEFAULTS.PLOT.ROBUST_PERCENTILE'}) -->
                if val=self.get(provided=UNSET)==True, use DEFAULTS.PLOT.ROBUST_PERCENTILE;
                if provided not UNSET, or val!=True, just use val.
            Also, in this case, stores the dict portion as defaults_keystring_dict.
    validation: None, list-like, or callable.
        if provided, tells how to validate any value provided to this property.
        list-like --> value must be in this list.
        callable --> ensure that bool(callable(value)) evaluates to True.
    validation_message: None or str.
        if provided, tells what to append to error message if validation fails;
        base error message will always be: "invalid value for {self.key!r}: {value!r}."
        If validation_message does not start with a space, one will be prepended.

    ntab: number of tabs to indent doc by, underneath typestr.
        default 1, is a good number for docstrings of module-level objects.
        (but, e.g. for methods defined inside a class, probably use ntab=2.)
    tab: string to use for a single tab.
    '''

    def __init__(self, key, default, typestr, doc, defaults_keystring=None,
                 validation=None, validation_message=None, *, ntab=1, tab=DEFAULTS.TAB):
        # essential info:
        self.key = key
        self.default = default
        self.typestr = typestr
        self.doc = inspect.cleandoc(doc)
        if isinstance(defaults_keystring, tuple):
            assert len(defaults_keystring)==2
            assert isinstance(defaults_keystring[1], dict)
            self.defaults_keystring = defaults_keystring[0]
            self.defaults_keystring_dict = defaults_keystring[1]
        else:
            self.defaults_keystring = defaults_keystring
            self.defaults_keystring_dict = None
        self.validation = validation
        self.validation_message = validation_message
        # formatting:
        self.ntab = ntab
        self.tab = tab

    def get(self, provided=UNSET):
        '''get the value from self. return provided if not UNSET, else return self.get_default()'''
        if provided is UNSET:
            return self.get_default()
        else:
            return provided

    def get_default(self):
        '''get the default value: self.default if defaults_keystring is None else value from DEFAULTS.'''
        if self.defaults_keystring is None:
            return self.default
        else:
            result = self.defaults_keystring_value()
            if (self.defaults_keystring_dict is not None) and (result in self.defaults_keystring_dict):
                result = self._value_from_DEFAULTS_str(self.defaults_keystring_dict[result],
                                                        s_name=f'self.defaults_keystring_dict[{result!r}]')
            return result

    def defaults_keystring_value(self):
        '''get the value from DEFAULTS.'''
        s = self.defaults_keystring
        return self._value_from_DEFAULTS_str(s, s_name='self.defaults_keystring')

    @staticmethod
    def _value_from_DEFAULTS_str(s, *, s_name=''):
        '''get any value from DEFAULTS. s_name used if raising an error message.'''
        assert s is not None, f"{s_name} must be provided in order to get value from defaults"
        assert s.startswith('DEFAULTS.'), f"{s_name} must start with 'DEFAULTS.'"
        s = s[len('DEFAULTS.'):]
        if s.startswith('PLOT.'):
            s = s[len('PLOT.'):]
            return getattr(DEFAULTS.PLOT, s)
        else:
            return getattr(DEFAULTS, s)

    def __iter__(self):
        '''returns iter of tuple of (key, default, typestr, doc, defaults_keystring, validation).
        Thus, can create new PlotSetting with:
            PlotSetting(self.key, *iter(self))
        '''
        return iter((self.default, self.typestr, self.doc, self.defaults_keystring, self.validation))

    # # # VALIDATION # # #
    def validate(self, value):
        '''ensures that value is valid for this setting. Raise InputError if it is not.
        return (if no InputError is raised) depends on self.validation:
            None --> return None
            iterable --> return True if valid in self.validation
            callable --> return self.validation(value)
        '''
        vv = self.validation
        if vv is None:
            return
        elif callable(vv):
            result = vv(value)
            if result:
                return result
            else:
                errmsg_add = self._get_validation_message()
                raise InputError(f'invalid value for {self.key!r}: {value!r}.{errmsg_add}')
        else:
            if value in vv:
                return True
            else:
                errmsg_add = self._get_validation_message(default=f'Expected one of: {vv!r}.')
                raise InputError(f'invalid value for {self.key!r}: {value!r}.{errmsg_add}')

    def _get_validation_message(self, default=None):
        '''return self.validation_message, with a space prepended if needed.
        if default provided and self.validation_message is None, use default.'''
        vmessage = self.validation_message
        if vmessage is None and default is not None:
            vmessage = default
        if (vmessage is not None) and (len(vmessage)>0) and (not vmessage[0].isspace()):
            vmessage = f' {vmessage}'  # prepend a space.
        return vmessage

    # # # DOCUMENTATION # # #
    def defaults_keystring_formatting(self):
        '''returns dict which will be used to .format the docstring.
        If self.defaults_keystring is provided:
            {'default_value': self.defaults_keystring_value(),
            'default_key': self.defaults_keystring,
            'default': f'{self.defaults_keystring} (default: {self.defaults_keystring_value()})'}
        Also, if self.defaults_keystring_dict is provided, include something similar for each key in dict,
            adding 'defaults_dict_{key}_value', 'defaults_dict_{key}_key', and 'defaults_dict_{key}'.
        E.g. if defaults_keystring=('DEFAULTS.PLOT.ROBUST', {True: 'DEFAULTS.PLOT.ROBUST_PERCENTILE'}),
            when DEFAULTS.PLOT.ROBUST = True and DEFAULTS.PLOT.ROBUST_PERCENTILE = 2.0:
                {'default_value': True
                'default_key': 'DEFAULTS.PLOT.ROBUST',
                'default': 'DEFAULTS.PLOT.ROBUST (default: True)',
                'defaults_dict_True_value': 2.0,
                'defaults_dict_True_key': 'DEFAULTS.PLOT.ROBUST_PERCENTILE',
                'defaults_dict_True': 'DEFAULTS.PLOT.ROBUST_PERCENTILE (default: 2.0)'}
        '''
        if self.defaults_keystring is None:
            result = dict()
        else:
            default_value = self.defaults_keystring_value()
            default_key = self.defaults_keystring
            result = {'default_value': default_value,
                        'default_key': default_key,
                        'default': f'{default_key} (default: {default_value})'}
        if getattr(self, 'defaults_keystring_dict', None) is not None:
            for key, value in self.defaults_keystring_dict.items():
                default_key = value
                default_value = self._value_from_DEFAULTS_str(value, s_name=f'self.defaults_keystring_dict[{key}]')
                result[f'defaults_dict_{key}_value'] = default_value
                result[f'defaults_dict_{key}_key'] = default_key
                result[f'defaults_dict_{key}'] = f'{default_key} (default: {default_value})'
        return result

    def docstring(self, ntab=UNSET):
        '''return the docstring for this setting, in format:
        {self.typestr} (default: {self.default})
            {self.doc}

        the typestr line will not be indented at all;
        the docstring line(s) will be indented by self.tab * ntab.

        self.doc will be hit by .format(**self.defaults_keystring_formatting()).
        (If self.defaults_keystring is None, this shouldn't change the result.)
        '''
        if ntab is UNSET: ntab = self.ntab
        indent = self.tab * (ntab + 1)
        s = f'{self.typestr} (default: {self.default})'
        s += f'\n{textwrap.indent(self.doc, indent)}'
        s = s.format(**self.defaults_keystring_formatting())
        return s

    def docs_dict(self, ntab=UNSET):
        '''returns {self.key: self.docstring(ntab)}'''
        return {self.key: self.docstring(ntab)}

    def __str__(self):
        return self.docstring()

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}({self.key!r}, default={self.default!r}, typestr={self.typestr!r})'


class SharedPlotSetting(PlotSetting):
    '''PlotSetting which is shared amongst multiple kwargs.
    E.g., in DEFAULT_PLOT_SETTINGS:
        'kw_subplots_adjust': [
            ('hspace', 'wspace', 'bottom', 'top', 'left', 'right'),
            (None, "None or number (probably between 0 and 1)",
            """corresponding value that will be used during plt.subplots_adjust(...).
            None --> use rcParams["figure.subplot.*"]."""),
            ],
        corresponds to 6 kwargs, all with the same default value and docstring.
        No need to repeat the docstring multiple times.

    self.docs_dict() will contain the value from super(),
        in addition to {docs_key: "{', '.join(self.subkeys)}: {docstring}"}.
        E.g. when doing PlotSettings.format_docstring(),
            formatted func could either contain individual keys, like:
                hspace: {hspace}
                wspace: {wspace}
                ...
            OR, just contain the docs_key, like:
                {kw_subplots_adjust}
    '''
    def __init__(self, *args_super, docs_key, shared_keys, **kw_super):
        super().__init__(*args_super, **kw_super)
        self.docs_key = docs_key
        self.shared_keys = shared_keys

    def _docs_key_docs_dict(self, ntab=UNSET):
        '''return dict with key self.docs_key, "{', '.join(self.subkeys)}: {self.docstring(ntab)}".'''
        return {self.docs_key: f'{", ".join(self.shared_keys)}: {self.docstring(ntab)}'}

    def docs_dict(self, ntab=UNSET):
        '''returns super().docs_di'''
        result = self._docs_key_docs_dict(ntab)
        super_docs_dict = super().docs_dict(ntab)
        result.update(super_docs_dict)
        return result

    def __repr__(self):
        contents = [f'{self.key!r}',
                    f'docs_key={self.docs_key!r}',
                    f'shared_keys={self.shared_keys!r}',
                    f'default={self.default!r}',
                    f'typestr={self.typestr!r}',]
        return f'{type(self).__name__}({", ".join(contents)})'


class PlotSettings():
    '''class for managing plot settings.
    class stores default settings and docstrings;
    instance stores current settings, e.g. for a particular object.

    UNSET is used as default value for many settings;
    this is to distinguish between "not provided" and "provided as None",
    since None is a valid value for many settings (usually, None means "use matplotlib defaults).

    all settings can be provided as kwargs during __init__.
    kwargs which are not recognized as valid PlotSettings.valid_keys() will be ignored.

    CAUTION: adjusting plot settings after instantiation might break mutually-dependent settings.
        [TODO] make mutually-dependent settings be property-like, instead, to mitigate this?^

    CAUTION: do not use self.kw directly;
        instead, use self.get(key), self.set(**kw), self[key], self[key]=value, or del self[key].
        This ensures values are validated and that only valid keys are used.
    '''
    DEFAULT_SETTINGS = DEFAULT_PLOT_SETTINGS
    MPL_KWARGS = MPL_KWARGS

    def __init__(self, *, pop_from=dict(), **kw):
        self.kw = self.dict_from_valid_keys(kw)
        self.update(pop_from, pop=True)

    # # # SET / GET # # #
    def set(self, **kw):
        '''set plot settings, in self.
        kwargs which are not recognized as valid PlotSettings.valid_keys() will be ignored.
        '''
        self.update(self.dict_from_valid_keys(kw))

    def get(self, key, provided=UNSET, *, default=NO_VALUE, fdefault=NO_VALUE, last_resort_default=NO_VALUE):
        '''get a plot setting. if provided is not UNSET, return it.
    
        There are multiple ways to set defaults.
        A pneumonic for the order of precedence when multiple defaults / values are entered, is:
            "provided > self.kw > default > fdefault > self.DEFAULT_SETTINGS > last_resort_default"

        More precisely, when getting key, does the following, in order:
            -- assert key in self.valid_keys(); raise PlotSettingsError if not.
            - return provided (if not UNSET).
            - return self.kw[key] (if it exists and is not UNSET)
            - return default (if not NO_VALUE).
            - return fdefault() (if not NO_VALUE).
            - return self.get_default(key), if key in self.DEFAULT_SETTINGS.
            - return last_resort_default, if it was provided (not NO_VALUE).
            -- give up; raise PlotSettingsError.
        '''
        # note: the defaults for kwargs to this function are NO_VALUE instead of UNSET,
        # to avoid ambiguity with possibly wanting to use UNSET as a default value...
        # however, provided must have a default of UNSET, to ease with usage like:
        #   def func(..., setting_name=UNSET, ...):
        #       setting_value = plot_settings..get('setting_name', setting_name, ...)
        #       ...
        if key not in self.valid_keys():
            raise PlotSettingsError(f'invalid key: {key!r}. Not found in self.valid_keys()')
        if provided is not UNSET:
            return provided
        result = self.kw.get(key, UNSET)
        if result is not UNSET:
            return result
        if default is not NO_VALUE:
            return default
        if fdefault is not NO_VALUE:
            return fdefault()
        if key in self.DEFAULT_SETTINGS:
            return self.get_default(key)
        if last_resort_default is not NO_VALUE:
            return last_resort_default
        # else, failed to get key.
        errmsg = (f'key missing: {key!r}. key is in {type(self).__name__}.valid_keys(), '
                    'but not set in self.kw, and not in self.DEFAULT_SETTINGS;\n'
                    "also did not enter value for 'provided', 'default', 'fdefault', or 'last_resort_default'.")
        raise PlotSettingsError(errmsg)

    @classmethod
    def default(cls, key):
        '''get the default for a plot setting.
        This is the default entered at the definition of the setting; it might be UNSET.
        See also cls.get_default.
        '''
        setting = cls.DEFAULT_SETTINGS[key]
        return setting.default

    @classmethod
    def get_default(cls, key):
        '''get the default value for a plot setting.
        This is setting.get_default(), where setting is cls.DEFAULT_SETTINGS[key].
        Note that setting.get_default() does not necessarily equal setting.default;
            in particular they will differ if the default is UNSET,
            and a defaults_keystring was provided to that setting.
        '''
        setting = cls.DEFAULT_SETTINGS[key]
        return setting.get_default()

    def is_default(self, key):
        '''return whether key has the default value for this setting.'''
        return key not in self.kw

    def __getitem__(self, key):
        '''return self.get(key).'''
        return self.get(key)

    def __setitem__(self, key, value):
        '''set self.kw[key] = value, after ensuring this is valid, via self.validate.
        Only set key if value is not self.default(key) or self.get_default(key) (compared via 'is').
        '''
        self.validate(key, value)
        if DEFAULTS.DEBUG>=1:
            self.kw[key] = value
        elif key not in self.DEFAULT_SETTINGS:
            self.kw[key] = value   # not in default settings; default & get_default would crash.
        elif value is not self.default(key) and value is not self.get_default(key):
            self.kw[key] = value

    def __delitem__(self, key):
        '''delete self.kw[key]'''
        del self.kw[key]

    def __contains__(self, key):
        '''returns whether key is in self.kw.'''
        return key in self.kw

    def update(self, d, *, pop=False):
        '''update self.kw with d, but only for keys in self.valid_keys().
        if pop, pop keys from d instead of just copying values.
        '''
        valid_keys = self.valid_keys()
        d_keys = tuple(d.keys())
        for key in d_keys:
            if key in valid_keys:
                value = d.pop(key) if pop else d[key]
                self.__setitem__(key, value)

    def take_from(self, d):
        '''update self.kw with d, popping keys from d instead of just copying values.'''
        self.update(d, pop=True)

    def copy(self):
        '''return new PlotSettings, with copy of self.kw'''
        return PlotSettings(**self.kw)

    # # # VALIDATION OF KEYS & VALUES # # #
    @classmethod
    def validate(cls, key, value):
        '''raise InputError if value is invalid for key, or PlotSettingsError if key not in self.valid_keys().
        return None if no validation was performed, else the result of validate (see PlotSetting.validate()).
        '''
        if key in cls.valid_keys():
            if key in cls.DEFAULT_SETTINGS:
                setting = cls.DEFAULT_SETTINGS[key]
                return setting.validate(value)
        else:
            raise PlotSettingsError(f'invalid key: {key!r}. Not found in self.valid_keys()')

    @classmethod
    def valid_keys(cls):
        '''return set of all valid keys for plot settings.
        This includes all keys from cls.DEFAULT_SETTINGS as well as all mpl keys.
        '''
        result = set(cls.DEFAULT_SETTINGS.keys())
        for mpl_funcname in cls.MPL_KWARGS:
            result.update(cls.get_mpl_keys(mpl_funcname, mode='flat'))
        return result

    @classmethod
    def dict_from_valid_keys(cls, d, *, pop=False):
        '''return dict but only with keys that are valid for PlotSettings, i.e. in PlotSettings.valid_keys().
        if pop, also pop these keys from d instead of just copying their values.
        '''
        result = dict()
        for key in cls.valid_keys():
            if key in d:
                if pop:
                    result[key] = d.pop(key)
                else:
                    result[key] = d[key]
        return result

    # # # DOCUMENTATION # # #
    @classmethod
    def docs_dict(cls, ntab=1):
        '''return dict of {key: docstring} for all settings. See: format_docstring.'''
        result = dict()
        for key, setting in cls.DEFAULT_SETTINGS.items():
            result.update(setting.docs_dict(ntab))
        return result

    @classmethod
    def format_docstring(cls, *, ntab=1, **kw_format_docstring):
        '''return function decorator which formats docstring of a plotter function.
        Provides docs for all settings, by default.
        ntab tells number of tabs for docs.
            Use 1 for module-level functions, 2 for class methods, 3+ if indenting deeper than that...
        additional kw go to tools.format_docstring.
        '''
        return format_docstring(**cls.docs_dict(ntab=ntab), **kw_format_docstring)
        
    # # # DEFAULTS # # #
    @classmethod
    def _instantiate_default_settings(cls):
        '''convert all DEFAULT_SETTINGS to PlotSetting (or SharedPlotSetting) instances.'''
        result = dict()
        for key, value in cls.DEFAULT_SETTINGS.items():
            if isinstance(value, tuple):
                result[key] = PlotSetting(key, *value)
            elif isinstance(value, list):
                shared_keys, v = value
                for subkey in shared_keys:
                    result[subkey] = SharedPlotSetting(subkey, *v, docs_key=key, shared_keys=shared_keys)
            else:
                raise TypeError(f'invalid type for DEFAULT_SETTINGS[{key}]: {type(value)}')
        cls.DEFAULT_SETTINGS = result

    # # # POP MATPLOTLIB KWARGS # # #
    @classmethod
    def cls_pop_mpl_kwargs(cls, mpl_funcname, kw, *, defaults=dict(), **other_kwargs):
        '''pops kwargs from kw which are supposed to go to matplotlib func with name mpl_funcname.
        valid mpl_funcname options determined by cls.MPL_KWARGS.

        returns dict of kwargs to pass to matplotlib func.

        if other_kwargs provided, use these for values not found in kw, if not UNSET.
        if defaults provided, use these for values not found in kw or other_kwargs, if not UNSET.

        Note: might be fancy, e.g. 'axes_class' from kw for 'plt.subplots'
            would be passed as subplot_kw=dict(axes_class=...).
        '''
        result = dict()
        for key in cls.get_mpl_keys(mpl_funcname):
            if isinstance(key, str):
                v = RESULT_MISSING
                if key in kw:
                    v = kw.pop(key)
                if v is RESULT_MISSING or v is UNSET:
                    if key in other_kwargs:
                        v = other_kwargs.pop(key)
                    if v is RESULT_MISSING or v is UNSET:
                        if key in defaults:
                            v = defaults[key]
                if v is not RESULT_MISSING and v is not UNSET:
                    result[key] = v
            elif isinstance(key, tuple):
                # nested kwargs. [TODO] encapsulate repeated code from above...
                key, nested_keys = key
                nested_kw = dict()
                for k in nested_keys:
                    v = RESULT_MISSING
                    if k in kw:
                        v = kw.pop(k)
                    if v is RESULT_MISSING or v is UNSET:
                        if k in other_kwargs:
                            v = other_kwargs.pop(k)
                        if v is RESULT_MISSING or v is UNSET:
                            if k in defaults:
                                v = defaults[k]
                    if v is not RESULT_MISSING and v is not UNSET:
                        nested_kw[k] = v
                if len(nested_kw) > 0:
                    result[key] = nested_kw
        return result

    @classmethod
    def cls_get_mpl_kwargs(cls, mpl_funcname, **kw):
        '''returns dict of kwargs to pass to matplotlib func with name mpl_funcname.
        see also: pop_mpl_kwargs.
        '''
        return cls.cls_pop_mpl_kwargs(mpl_funcname, kw)   # kw was passed with ** so don't need to make a .copy()

    @classmethod
    def get_mpl_keys(cls, mpl_funcname, *, mode='nested'):
        '''returns tuple of keys which are supposed to go to matplotlib func with name mpl_funcname, as kwargs.

        some elements might be tuples instead of strings. If so, it means that they are nested kwargs.
            E.g. ('subplot_kw', ('axes_class', 'polar')) means that kw['axes_class'] and kw['polar']
            would be passed to mpl func as subplot_kw=dict(axes_class=kw['axes_class'], polar=kw['polar']).

        if mode=='flat', returns a flat tuple of keys, instead of nesting them.
            in this case, the dict-like matplotlib keys will NOT be included.
            E.g. in the example above, the result would be ('axes_class', 'polar').
            ('subplot_kw' gets excluded since it is a dict-like key.)
        '''
        if mpl_funcname not in cls.MPL_KWARGS:
            raise InputError(f'invalid mpl_funcname: {mpl_funcname!r}.')
        settings = cls.MPL_KWARGS[mpl_funcname].copy()
        result = []
        here = settings.pop('here', ())
        result.extend(here)
        from_ = settings.pop('from', ())
        for f in from_:
            result.extend(cls.get_mpl_keys(f, mode=mode))
        # other keys mean "pass these kwargs as this key=dict(other kwargs)".
        for key, f in settings.items():
            # [TODO] encapsulate repeated code from above...
            keys_from_key = []
            fcopy = f.copy()
            here = fcopy.pop('here', ())
            keys_from_key.extend(here)
            from_ = fcopy.pop('from', ())
            for g in from_:
                keys_from_key.extend(cls.get_mpl_keys(g, mode=mode))
            # put keys_from_key into result.
            if mode=='nested':
                result.append((key, tuple(keys_from_key)))
            elif mode=='flat':
                result.extend(keys_from_key)
            else:
                raise InputError(f'invalid mode: {mode!r}.')
        return tuple(result)

    def get_mpl_func_settings(self, mpl_funcname):
        '''return dict of {key: value} for all mpl kwargs for mpl_funcname, with values from self.
        Note: this is different get_mpl_kwargs, since the result will not have any nested kwargs.
        '''
        result = dict()
        for key in self.get_mpl_keys(mpl_funcname, mode='flat'):
            val = self.get(key, last_resort_default=UNSET)
            if val is not UNSET:
                result[key] = val
        return result

    def pop_mpl_kwargs(self, mpl_funcname, kw, **other_kwargs):
        '''return kwargs to use for mpl_funcname, using values from self as defaults, and kw if provided.
        pop values from kw if they are used.

        use values from kw if possible,
        else from other_kwargs if possible (exists and not UNSET).
        else from self if possible (not UNSET).
        '''
        defaults = self.get_mpl_func_settings(mpl_funcname)
        return self.cls_pop_mpl_kwargs(mpl_funcname, kw, defaults=defaults, **other_kwargs)

    def get_mpl_kwargs(self, mpl_funcname, **kw):
        '''return kwargs to use for mpl_funcname, using values from self as defaults, and kw if provided.'''
        return self.pop_mpl_kwargs(mpl_funcname, kw)

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}({self.kw!r})'


PlotSettings._instantiate_default_settings()


class PlotSettingsMixin():
    '''mixin which provides obj.plot_settings, a PlotSettings object.
    Additionally, kwargs in __init__ will be passed to obj.plot_settings,
        which will remember all valid plot settings entered into it.

    Be sure to put PlotSettingsMixin class ahead of matplotlib classes in MRO,
        so that PlotSettingsMixin can properly pop plot settings kwargs.

    mpl_super: list of strings, names of matplotlib super() classes (from MPL_KWARGS.keys()).
        If provided, pass kwargs for those classes, from self.plot_settings, during super().__int__.
    '''
    def __init__(self, *args_super, mpl_super=[], **kw):
        self.plot_settings = PlotSettings(pop_from=kw)  # pop plot settings kwargs from kw.
        if len(mpl_super) > 0:
            mpl_kwargs = dict()
            for mpl_funcname in mpl_super:
                mpl_kwargs.update(self.plot_settings.get_mpl_kwargs(mpl_funcname))
            kw.update(mpl_kwargs)
        super().__init__(*args_super, **kw)

    def __init_subclass__(cls, *args_super, **kw):
        '''appends note about using self.plot_settings, to cls.__doc__.
        if "PlotSettings" or "plot_settings" appears in cls.__doc__, do NOT append this note;
            assuming instead that this means the doc already mentions how to use plot_settings.
        '''
        msg = (f'To view or adjust plot settings in self, see self.plot_settings, or help(self.plot_settings).\n')
        msg = textwrap.indent(msg, DEFAULTS.TAB)
        if cls.__doc__ is None:
            cls.__doc__ = msg
        elif 'PlotSettings' not in cls.__doc__ and 'plot_settings' not in cls.__doc__:
            cls.__doc__ += f'\n\n{msg}'



"""
File Purpose: reading genrad tables, e.g. radtab.dat files.
These include lookup tables which might improve ion densities compared to saha.
"""
import os

import numpy as np
import xarray as xr

from .mhd_fluids import ElementHaver
from ...defaults import DEFAULTS
from ...errors import (
    InputError,
    FileContentsError,
    FluidKeyError,
)
from ...tools import alias

# location of the folder in PlasmaCalcs containing default genrad tables, on this machine.
# this is used when trying to get one of the default genrad tables.
# __file__ means the file where this line of code is located.
DEFAULTS.GENRAD_TAB_DIR = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'mhd_tables')

def _float(s):
    '''convert fortran float string to np.float64.'''
    return np.float64(s.replace('d', 'e'))

def _consume_comment_lines(enumlines, *, start=None, comment=None):
    r'''eat (i, line) from enumlines until reaching a nonempty line that doesn't start with '*'.
    return (next(enumlines), comment.strip()) where next(enumlines) is the next line.

    start: None or (i, line)
        if provided, start here. Otherwise, start at next(enumlines)
    comment: None or str
        if provided, append result to this comment + '\n'.
    '''
    comment = '' if comment is None else (comment + '\n')
    if start is None:
        start = next(enumlines)
    i, line = start
    while line.startswith('*') or len(line)==0:
        if len(line) != 0:
            comment += line[len('*'):].strip() + '\n'
        i, line = next(enumlines)
    return ((i, line), comment.strip())

def _consume_data_lines(enumlines, *, start=None, Nnums=None, var=None):
    '''eat (i, line) from enumlines until reaching a line that startswith ' or *
    treat these lines as data; split() and convert from fortran float strs to np.float64.
    return (next(enumlines), data), where next(enumlines) is the next line.
    
    start: None or (i, line)
        if provided, start here. Otherwise, start at next(enumlines)
    Nnums: None or int
        if provided, ensure len(data)==N_nums, else crash with FileContentsError.
    var: None or str
        if provided, included this name in errmsg if crashing with FileContentsError.
    '''
    datalines = []
    if start is None:
        start = next(enumlines)
    i, line = start
    while (not line.startswith("'")) and (not line.startswith('*')):
        datalines.append(line)
        i, line = next(enumlines)
    data_str = ' '.join(datalines).split()
    data = [_float(num) for num in data_str]  # fortran float sometimes uses d instead of e
    if (Nnums is not None) and (len(data) != Nnums):
        errmsg = f'got {len(data)} numbers, expected {Nnums}, near line {i}'
        if var is not None:
            errmsg = errmsg + f', for var={var!r}'
        raise FileContentsError(errmsg)
    return ((i, line), data)

def _consume_indepvar_data_lines(enumlines, *, start=None, var=None):
    '''eat (i, line) from enumlines until reaching a line that startswith ' or *
    expect first line to be an int, and remaining lines to contain that many values.
    treat remaining lines as data; split() and convert from fortran float strs to np.float64.
    return (next(enumlines), data), where next(enumlines) is the next line.
    
    start: None or (i, line)
        if provided, start here. Otherwise, start at next(enumlines)
    var: None or str
        if provided, included this name in errmsg if crashing with FileContentsError.
    '''
    if start is None:
        start = next(enumlines)
    i, line = start
    try:
        Nnums = int(line)
    except ValueError:
        errmsg = f'expected integer on line {i}'
        if var is not None: errmsg += f', for var={var!r}'
        errmsg += f', got line={line!r}'
        raise FileContentsError(errmsg)
    return _consume_data_lines(enumlines, start=next(enumlines), Nnums=Nnums, var=var)


class GenradTable():
    '''single table from genrad file, with some var as a function of T, tauh, or column mass.
    
    indict: dict with length 1
        key = independent variable name. Probably 'temp', 'tauh', or 'cmass'.
        indict[key] = independent variable data
    outdict: dict with length 1
        key = dependent variable name
        outdict[key] = dependent variable data
    coords: None or dict of scalars
        if provided, assign these coords to self.xtable.
    comment: None or str
        if provided, any arbitrary comment about this variable.
    
    self will also store self.xtable, an xarray.DataArray, with
        name = outdict key
        values = outdict values
        coords = {xtable[indict key]: indict values}
    '''
    def __init__(self, indict, outdict, *, coords=None, comment=None):
        self.indict = indict
        self.outdict = outdict
        self.coords = coords
        self.comment = comment
        self.init_checks()
        self.init_xtable()
        
    def init_checks(self):
        if not isinstance(self.indict, dict):
            raise InputError('expected indict to be a dict')
        if not isinstance(self.outdict, dict):
            raise InputError('expected outdict to be a dict')
        if len(self.indict) != 1:
            raise InputError('expected indict to have length 1')
        if len(self.outdict) != 1:
            raise InputError('expected outdict to have length 1')

    invar = property(lambda self: list(self.indict.keys())[0], doc='independent variable name')
    outvar = property(lambda self: list(self.outdict.keys())[0], doc='dependent variable name')
    invals = property(lambda self: np.asanyarray(self.indict[self.invar]), doc='independent variable values')
    outvals = property(lambda self: np.asanyarray(self.outdict[self.outvar]), doc='dependent variable values')
    
    def init_xtable(self):
        '''initialize self.xtable.'''
        kw = dict(
            dims=[self.invar],
            coords={self.invar: self.invals},
            name=self.outvar,
        )
        if self.coords is not None:
            kw['coords'] = {**kw['coords'], **self.coords}
        if self.comment is not None:
            kw['attrs'] = {'comment': self.comment}
        self.xtable = xr.DataArray(self.outvals, **kw)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'{self.invar!r}', f'{self.outvar!r}', f'len={self.invals.size}']
        if self.comment is not None:
            comment = self.comment.split('\n')
            comminfo = f'comment={comment[0][:40]!r}'
            if len(comment[0]) > 40 or len(comment) > 1:
                comminfo += '...'
            contents.append(comminfo)
        return f'{type(self).__name__}({", ".join(contents)})'

    # # # INTERPOLATING # # #
    __call__ = alias('interp')

    def interp(self, invalues, *, fill_value='extrapolate', **kw_xarray_interp):
        '''interpolate from self.xtable.interp(**kw_interp).
        invalues: values of independent variable. Same units as self.invals.
        '''
        kwargs = None if fill_value is None else dict(fill_value=fill_value)
        kw = dict(kwargs=kwargs, **kw_xarray_interp)
        return self.xtable.interp({self.invar: invalues}, **kw)


class GenradTableManager(dict):
    '''manages genrad tables. Dict of {quantity (str): GenradTable}.

    thresh: 2-tuple of (values, comment)
        thresh values & comment from genrad file. [TODO] what do they mean?
        internally stored as (self.thresh, self.threshcomment)
    incrad: 2-tuple of (value, comment)
        incrad values & comment from genrad file. [TODO] what dooes it mean?
        internally stored as (self.incrad, self.incradcomment)

    self.filename will tell abspath to file if created self via GenradTableManager.from_file()
    '''
    genrad_table_cls = GenradTable
    INDEP_VARS = ['temp', 'tauh', 'cmass']  # <-- treat these as independent variables.

    def __init__(self, *args_dict, thresh=(None, None), incrad=(None, None), **kw_dict):
        (self.thresh, self.threshcomment) = thresh
        (self.incrad, self.incradcomment) = incrad
        super().__init__(*args_dict, **kw_dict)

    @classmethod
    def from_defaults(cls, name='radtab'):
        '''return GenradTableManager from one of the default genrad table files.
        name = one of cls.DEFAULTS.keys().
        '''
        try:
            filename = cls.DEFAULTS[name]
        except KeyError:
            raise InputError(f'name={name!r}; expected one of {tuple(cls.DEFAULTS.keys())}.') from None
        filepath = os.path.join(DEFAULTS.GENRAD_TAB_DIR, filename)
        return cls.from_file(filepath)

    DEFAULTS = {  # dict of {shorthand name: filename}
        'radtab': 'radtab.dat',
    }

    @classmethod
    def from_file(cls, genrad_file):
        '''load genrad tables from file.'''
        with open(genrad_file, 'r') as f:
            lines = f.readlines()
        lines = [line.strip() for line in lines]

        result = {}
        kw_result = {}
        enumlines = enumerate(lines)
        # get header
        (i, line), header = _consume_comment_lines(enumlines)
        # get 'thin' data
        if line.strip() != "'temp'":  # expect first independent var=='temp'
            raise NotImplementedError(f"reading genrad file when first non-comment line (i={i}; line={line!r}) isn't 'temp'.")
        (i, line), tempdata = _consume_indepvar_data_lines(enumlines, var='temp')
        (i, line), thincomment = _consume_comment_lines(enumlines, start=(i, line))
        if line.strip() != "'thin'":  # expect first dependent var=='thin'
            raise NotImplementedError(f"reading genrad file when first dependent var (i={i}; line={line!r}) isn't 'thin'.")
        (i, line), thincomment = _consume_comment_lines(enumlines, comment=thincomment)
        thinunits = _float(line)  # at least, I think this is units... anyways, there's a single number before the data.
        (i, line), thindata = _consume_data_lines(enumlines, Nnums=len(tempdata), var='thin')
        result['thin'] = GenradTable({'temp': tempdata}, {'thin': thindata}, coords={'units?': thinunits}, comment=thincomment)
        # get 'thresh' values
        (i, line), threshcomment = _consume_comment_lines(enumlines, start=(i, line))
        if line.strip() != "'thresh'":  # expect var=='thresh'
            raise NotImplementedError(f"reading genrad file when var after 'thin' is not 'thresh'. Got i={i}; line={line!r}.")
        (i, line), threshcomment = _consume_comment_lines(enumlines, comment=threshcomment)
        (i, line), threshdata = _consume_data_lines(enumlines, start=(i, line), Nnums=2, var='thresh')  # tmin, tscl
        kw_result['thresh'] = (threshdata, threshcomment)
        # get 'incrad' value
        (i, line), incradcomment = _consume_comment_lines(enumlines, start=(i, line))
        if line.strip() != "'incrad'":  # expect var=='incrad'
            raise NotImplementedError(f"reading genrad file when var after 'thresh' is not 'incrad'. Got i={i}; line={line!r}.")
        (i, line), incradcomment = _consume_comment_lines(enumlines, comment=incradcomment)
        (i, line), incraddata = _consume_data_lines(enumlines, start=(i, line), Nnums=1, var='incrad')
        kw_result['incrad'] = (incraddata[0], incradcomment)
        # get first independent variable
        (i, line), comment = _consume_comment_lines(enumlines, start=(i, line))
        indepvar = line.strip().strip("'")
        if indepvar not in cls.INDEP_VARS:
            raise FileContentsError(f"expected one of the INDEP_VARS {cls.INDEP_VARS}. Got i={i}; line={line!r}.")
        (i, line), indepdata = _consume_indepvar_data_lines(enumlines, var=indepvar)
        # get first dependent variable
        (i, line), comment = _consume_comment_lines(enumlines, start=(i, line), comment=comment)
        depvar = line.strip().strip("'")
        (i, line), depdata = _consume_data_lines(enumlines, var=depvar, Nnums=len(indepdata))
        result[depvar] = GenradTable({indepvar: indepdata}, {depvar: depdata}, comment=comment)

        # the rest of this file has a bunch of variables then data. Sometimes multiple dependent vars for same independent.
        (i, line), comment = _consume_comment_lines(enumlines, start=(i, line))
        var = line.strip().strip("'")
        while var != 'end':
            if var in cls.INDEP_VARS:
                indepvar = var
                (i, line), indepdata = _consume_indepvar_data_lines(enumlines, var=indepvar)
            else:
                depvar = var
                (i, line), depdata = _consume_data_lines(enumlines, var=depvar, Nnums=len(indepdata))
                result[depvar] = GenradTable({indepvar: indepdata}, {depvar: depdata}, comment=comment)
                comment = ''  # reset the comment to empty string after each dependent var.
            (i, line), comment = _consume_comment_lines(enumlines, start=(i, line), comment=comment)
            var = line.strip().strip("'")

        result = cls(result, **kw_result)
        result.filename = os.path.abspath(genrad_file)
        return result

    # # # DISPLAY # # #
    def __repr__(self):
        if len(self) <= 1:
            contents_str = super().__repr__()
        else:
            contents = [f'{key!r}: {val}' for key, val in self.items()]
            contents_str = '{\n    ' + ',\n    '.join(contents) + '\n}'
        return f'{type(self).__name__}({contents_str})'

    # # # INTERPOLATING # # #
    __call__ = alias('interp')

    def interp(self, var, invalues, **kw_interp):
        '''interpolate var from self[var].interp(values, **kw_interp).'''
        return self[var].interp(invalues, **kw_interp)

    # # # CHECKING AVAILABLE SPECIES # # #
    def neufrac_available(self, elem=None):
        '''return dict of {element name: neufrac variable}
        such that self[neufrac variable] = table telling neutral fraction for this element,
            as a funtion of temperature.

        elem: None, str, or ElementHaver (e.g. Element or Specie)
            if provided, return neufrac variable for elem (if str, else elem.get_element()).
            if no neufrac variable exists for this elem, return None.
        '''
        result = {}
        if 'h_ion' in self:
            result['H'] = 'h_ion'
        if 'mg_ion' in self:
            result['Mg'] = 'mg_ion'
        if 'ca_ion' in self:
            result['Ca'] = 'ca_ion'
        # [TODO] ^^ other elements must be added here in order for self('ionfrac_radtab') to know about them.
        if elem is None:
            return result
        if isinstance(elem, ElementHaver):
            elem = str(elem.get_element())
        return result.get(elem, None)

    def neufrac_table(self, elem):
        '''return self[var] for var telling neufrac for this elem.
        elem: str or ElementHaver (e.g. Element or Specie)
            element name, element, or object with elem.get_element().
        '''
        neufrac_var = self.neufrac_available(elem)
        if neufrac_var is None:
            raise FluidKeyError(f'no neutral fraction table available for elem={elem!r}.')
        return self[neufrac_var]



"""
File Purpose: base quantities for BifrostCalculator
"""

from ...mhd import MhdBasesLoader
from ...tools import simple_property


''' --------------------- BifrostBasesLoader --------------------- '''

class BifrostBasesLoader(MhdBasesLoader):
    '''base quantities based on Bifrost output.'''
    
    # # # DIRECTLY FROM BIFROST # # #
    @known_var(dims=['snap'])
    def get_r(self):
        '''mass density. (directly from Bifrost)
        assumes single fluid mode, i.e. result corresponds to the single fluid from Bifrost.
        '''
        self.assert_single_fluid_mode('r')
        return self.load_maindims_var_across_dims('r', u='mass_density', dims=['snap'])

    @known_var(dims=['snap', 'component'])
    def get_p(self):
        '''momentum density. (directly from Bifrost)
        assumes single fluid mode, i.e. result corresponds to the single fluid from Bifrost.
        '''
        self.assert_single_fluid_mode('p')
        return self.load_maindims_var_across_dims('p', u='momentum_density', dims=['snap', 'component'])

    @known_var(dims=['snap'])
    def get_e(self):
        '''energy density. (directly from Bifrost)
        Per unit volume, e.g. the SI units would be Joules / meter^3.
        assumes single fluid mode, i.e. result corresponds to the single fluid from Bifrost.
        '''
        self.assert_single_fluid_mode('e')
        return self.load_maindims_var_across_dims('e', u='energy_density', dims=['snap'])

    @known_var(dims=['snap', 'component'])
    def get_B(self):
        '''magnetic field. (directly from Bifrost)'''
        return self.load_maindims_var_across_dims('b', u='b', dims=['snap', 'component'])

    @known_var
    def get_gamma(self):
        '''adiabatic index. (directly from Bifrost)'''
        return self.params['gamma']

    # # # DERIVED # # #
    @known_var(deps=['p', 'r'])
    def get_u(self):
        '''velocity. u = p / r = (momentum density / mass density)'''
        return self('p') / self('r')

    @known_var(deps=['curl_B'])
    def get_J(self):
        '''current density. J = curl(B) / mu0.
        Per unit area, e.g. the SI units would be Amperes / meter^2.
        '''
        if self.J_stagger:
            with self.using(stagger_direct=False):  # don't stagger B within load_direct
                curl_B = self('centered_facecurl_B')  # curl of B, staggered to cell centers
        else:
            curl_B = self('curl_B')
        return curl_B / self.u('mu0')

    cls_behavior_attrs.register('J_stagger', default=True)
    J_stagger = simple_property('_J_stagger', default=True,
        doc='''whether to use stagger grid for derivatives during J = curl(B) / mu0.
        False --> apply naive derivatives (e.g. xarray differentiate) to stagger-centered B.
        True --> load staggered B across full grid, apply staggered derivatives,
                center result to cell centers, then apply any slices.''')

    # # # MISC TABIN VARS (other than EOS vars ne, T, and P) # # #
    @known_var(deps=['SF_e', 'SF_r'])
    def get_kappaR(self):
        '''Rosseland opacity, from plugging r and e into eos tables (see self.tabin).'''
        return self.get_ertab_var('kappaR', 'opacity')
